from __future__ import annotations

import random
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Protocol, TypeAlias

import msgspec

from ..game_modes import GameMode
from ..paths import default_runtime_dir
from ..pause_background import PauseBackground
from ..quest_level import QuestLevel
from ..render.rtx.mode import RtxRenderMode


def _default_rtx_render_mode() -> RtxRenderMode:
    return RtxRenderMode.CLASSIC

if TYPE_CHECKING:
    from grim.assets import LogoAssets, PaqTextureCache
    from grim.audio import AudioState
    from grim.config import CrimsonConfig
    from grim.console import ConsoleState
    from grim.geom import Vec2
    from grim.terrain_render import GroundRenderer

    from ..modes.quest_mode import QuestRunOutcome
    from ..net.lockstep_runtime import LockstepRuntime
    from ..net.rollback_runtime import RollbackRuntime
    from ..persistence.save_status import GameStatus


class GameConfig(msgspec.Struct, frozen=True):
    base_dir: Path = msgspec.field(default_factory=default_runtime_dir)
    assets_dir: Path | None = None
    width: int | None = None
    height: int | None = None
    fps: int = 60
    seed: int | None = None
    demo_enabled: bool = False
    no_intro: bool = False
    debug: bool = False
    rtx: bool = False
    preserve_bugs: bool = False
    pending_network_session: "PendingNetworkSession | None" = None


NetworkSessionMode = Literal["survival", "rush", "quests"]
NetworkSessionRole = Literal["host", "join"]
NetcodeMode = Literal["rollback", "lockstep"]


class LockstepEndpoint(msgspec.Struct, frozen=True):
    bind_host: str = "0.0.0.0"
    host: str = "127.0.0.1"
    port: int = 31993


class RollbackEndpoint(msgspec.Struct, frozen=True):
    relay_host: str = "127.0.0.1"
    relay_port: int = 31993
    room_code: str = ""


NetworkEndpoint: TypeAlias = LockstepEndpoint | RollbackEndpoint


class LockstepSessionConfig(msgspec.Struct, frozen=True):
    mode: NetworkSessionMode
    endpoint: LockstepEndpoint
    netcode_mode: Literal["lockstep"] = "lockstep"
    player_count: int = 1
    quest_level: str = ""
    rollback_max_ticks: int = 8
    reconnect_timeout_ms: int = 15_000
    input_delay_ticks: int = 1
    preserve_bugs: bool = False

    @property
    def quest_level_value(self) -> QuestLevel | None:
        return QuestLevel.try_parse(self.quest_level)


class RollbackSessionConfig(msgspec.Struct, frozen=True):
    mode: NetworkSessionMode
    endpoint: RollbackEndpoint
    netcode_mode: Literal["rollback"] = "rollback"
    player_count: int = 1
    quest_level: str = ""
    rollback_max_ticks: int = 8
    reconnect_timeout_ms: int = 15_000
    input_delay_ticks: int = 1
    preserve_bugs: bool = False

    @property
    def quest_level_value(self) -> QuestLevel | None:
        return QuestLevel.try_parse(self.quest_level)


NetworkSessionConfig: TypeAlias = LockstepSessionConfig | RollbackSessionConfig


class PendingNetworkSession(msgspec.Struct):
    role: NetworkSessionRole
    config: NetworkSessionConfig
    auto_start: bool = False
    started: bool = False
    error: str = ""


class HighScoresRequest(msgspec.Struct):
    game_mode_id: GameMode
    quest_stage_major: int = 0
    quest_stage_minor: int = 0
    highlight_rank: int | None = None


class FrontView(Protocol):
    def open(self) -> None: ...

    def close(self) -> None: ...

    def update(self, dt: float) -> None: ...

    def draw(self) -> None: ...

    def take_action(self) -> str | None: ...


class GameState(msgspec.Struct):
    base_dir: Path
    assets_dir: Path
    rng: random.Random
    config: CrimsonConfig
    status: GameStatus
    console: ConsoleState
    demo_enabled: bool
    preserve_bugs: bool
    logos: LogoAssets | None
    texture_cache: PaqTextureCache | None
    audio: AudioState | None
    resource_paq: Path
    session_start: float
    rtx_mode: RtxRenderMode = msgspec.field(default_factory=_default_rtx_render_mode)
    skip_intro: bool = False
    gamma_ramp: float = 1.0
    snd_freq_adjustment_enabled: bool = True
    menu_ground: GroundRenderer | None = None
    menu_ground_camera: Vec2 | None = None
    menu_sign_locked: bool = False
    stats_menu_easter_egg_roll: int = -1
    pause_background: PauseBackground | None = None
    pending_network_session: PendingNetworkSession | None = None
    network_runtime: "RollbackRuntime | LockstepRuntime | None" = None
    network_in_lobby: bool = False
    network_waiting_for_players: bool = False
    network_expected_players: int = 1
    network_connected_players: int = 1
    network_desync_count: int = 0
    network_resync_failure_count: int = 0
    network_last_error: str = ""
    pending_quest_level: str | None = None
    pending_high_scores: HighScoresRequest | None = None
    quest_outcome: QuestRunOutcome | None = None
    quest_fail_retry_count: int = 0
    terrain_regenerate_requested: bool = False
    survival_elapsed_ms: float = 0.0
    demo_trial_elapsed_ms: int = 0
    quit_requested: bool = False
    screen_fade_alpha: float = 0.0
    screen_fade_ramp: bool = False
    runtime_updates_per_frame: int = 0
    input_stall_count: int = 0
    ticks_advanced_per_frame: int = 0
    sim_ms: float = 0.0
    presentation_plan_ms: float = 0.0
    presentation_apply_ms: float = 0.0

    @property
    def pending_quest_level_value(self) -> QuestLevel | None:
        return QuestLevel.try_parse(self.pending_quest_level)

    @pending_quest_level_value.setter
    def pending_quest_level_value(self, value: QuestLevel | None) -> None:
        self.pending_quest_level = value.to_string() if value is not None else None


__all__ = [
    "FrontView",
    "GameConfig",
    "GameState",
    "HighScoresRequest",
    "LockstepEndpoint",
    "LockstepSessionConfig",
    "NetcodeMode",
    "NetworkEndpoint",
    "NetworkSessionConfig",
    "NetworkSessionMode",
    "PendingNetworkSession",
    "PauseBackground",
    "RollbackEndpoint",
    "RollbackSessionConfig",
]



from __future__ import annotations

import time
from collections.abc import Callable

import msgspec

from ..effects import FxQueue, FxQueueRotated
from ..game_modes import GameMode
from ..math_parity import f32
from ..perks.availability import perks_rebuild_available
from ..weapon_runtime import weapon_refresh_available
from .input import PlayerInput
from .input_frame import normalize_input_frame
from .presentation_step import PresentationStepCommands, plan_world_presentation_step
from .timing import FrameTiming
from .world_state import WorldEvents, WorldState


class PresentationRngTrace(msgspec.Struct):
    draws_total: int = 0
    draws_by_consumer: dict[str, int] = msgspec.field(default_factory=dict)


class DeterministicStepResult(msgspec.Struct):
    dt_sim: float
    timing: FrameTiming
    events: WorldEvents
    presentation: PresentationStepCommands
    presentation_plan_ms: float
    presentation_rng_trace: PresentationRngTrace
    post_apply_sfx_keys: tuple[str, ...] = ()


class StepPipelineOptions(msgspec.Struct, frozen=True):
    world_size: float
    damage_scale_by_type: dict[int, float]
    detail_preset: int
    gore_disabled: int
    auto_pick_perks: bool
    game_mode: GameMode
    demo_mode_active: bool
    perk_progression_enabled: bool
    game_tune_started: bool


def time_scale_reflex_boost_bonus(
    *,
    reflex_boost_timer: float,
    time_scale_active: bool,
    dt: float,
) -> float:
    """Apply Reflex Boost time scaling, matching the classic frame loop latch semantics."""

    # Native stores frame delta time in float32 (`frame_dt`). Many downstream systems
    # multiply `frame_dt` before rounding back to float32, so the *input* precision
    # matters even when Reflex Boost is inactive.
    dt_f32 = f32(float(dt))
    if not (float(dt_f32) > 0.0):
        return float(dt_f32)
    if not time_scale_active:
        return float(dt_f32)

    time_scale_factor = time_scale_reflex_boost_factor(
        reflex_boost_timer=float(reflex_boost_timer),
        time_scale_active=bool(time_scale_active),
    )
    return float(f32(float(dt_f32) * float(time_scale_factor)))


def time_scale_reflex_boost_factor(
    *,
    reflex_boost_timer: float,
    time_scale_active: bool,
) -> float:
    if not bool(time_scale_active):
        return 1.0
    reflex_f32 = f32(float(reflex_boost_timer))
    time_scale_factor = f32(0.3)
    if float(reflex_f32) < 1.0:
        time_scale_factor = f32((1.0 - float(reflex_f32)) * 0.7 + 0.3)
    return float(time_scale_factor)


def run_deterministic_step(
    *,
    world: WorldState,
    timing: FrameTiming,
    options: StepPipelineOptions,
    apply_world_dt_steps: bool = True,
    inputs: list[PlayerInput] | None,
    fx_queue: FxQueue,
    fx_queue_rotated: FxQueueRotated,
    defer_camera_shake_update: bool = False,
    defer_freeze_corpse_fx: bool = False,
    mid_step_hook: Callable[[], None] | None = None,
    rng_marks_out: dict[str, int] | None = None,
    trace_presentation_rng: bool = False,
) -> DeterministicStepResult:
    state = world.state
    rand = state.rng.rand

    def _mark(name: str) -> None:
        if rng_marks_out is None:
            return
        rng_marks_out[str(name)] = int(state.rng.state)

    inputs = normalize_input_frame(inputs, player_count=len(world.players)).as_list()

    _mark("gw_begin")
    state.game_mode = options.game_mode
    state.demo_mode_active = bool(options.demo_mode_active)

    weapon_refresh_available(state)
    _mark("gw_after_weapon_refresh")
    perks_rebuild_available(state)
    _mark("gw_after_perks_rebuild")
    _mark("gw_after_time_scale")

    prev_audio = [(player.shot_seq, player.weapon.reload_active, player.weapon.reload_timer) for player in world.players]
    prev_perk_pending = int(state.perk_selection.pending_count)

    events = world.step(
        float(timing.dt_sim),
        apply_world_dt_steps=apply_world_dt_steps,
        dt_player_local=float(timing.dt_player_local),
        defer_camera_shake_update=defer_camera_shake_update,
        defer_freeze_corpse_fx=defer_freeze_corpse_fx,
        mid_step_hook=mid_step_hook,
        inputs=inputs,
        world_size=float(options.world_size),
        damage_scale_by_type=options.damage_scale_by_type,
        detail_preset=int(options.detail_preset),
        gore_disabled=int(options.gore_disabled),
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        auto_pick_perks=bool(options.auto_pick_perks),
        game_mode=options.game_mode,
        perk_progression_enabled=bool(options.perk_progression_enabled),
        game_tune_started=bool(options.game_tune_started),
        rng_marks=rng_marks_out,
    )

    trace = PresentationRngTrace()

    def _rand_for(label: str):
        if not trace_presentation_rng:
            return rand

        def _draw() -> int:
            value = int(rand())
            trace.draws_total += 1
            trace.draws_by_consumer[str(label)] = int(trace.draws_by_consumer.get(str(label), 0)) + 1
            return value

        return _draw

    plan_ns_start = time.perf_counter_ns()
    presentation = plan_world_presentation_step(
        state=state,
        players=world.players,
        fx_queue=fx_queue,
        hits=events.hits,
        deaths=events.deaths,
        pickups=events.pickups,
        event_sfx=events.sfx,
        prev_audio=prev_audio,
        prev_perk_pending=int(prev_perk_pending),
        game_mode=options.game_mode,
        demo_mode_active=bool(options.demo_mode_active),
        perk_progression_enabled=bool(options.perk_progression_enabled),
        rand=rand,
        rand_for=_rand_for if trace_presentation_rng else None,
        detail_preset=int(options.detail_preset),
        gore_disabled=int(options.gore_disabled),
        game_tune_started=bool(options.game_tune_started),
        trigger_game_tune=events.trigger_game_tune,
        hit_sfx=events.hit_sfx,
        death_sfx_preplanned=events.death_sfx_preplanned,
    )
    presentation_plan_ms = (time.perf_counter_ns() - plan_ns_start) / 1_000_000.0

    if rng_marks_out is not None and trace_presentation_rng:
        rng_marks_out["ps_draws_total"] = int(trace.draws_total)
        for key, value in sorted(trace.draws_by_consumer.items()):
            rng_marks_out[f"ps_draws_{key}"] = int(value)

    return DeterministicStepResult(
        dt_sim=float(timing.dt_sim),
        timing=timing,
        events=events,
        presentation=presentation,
        presentation_plan_ms=float(presentation_plan_ms),
        presentation_rng_trace=trace,
    )



from __future__ import annotations

from enum import IntEnum

CREATURE_LIFECYCLE_ALIVE = 16.0
CREATURE_LIFECYCLE_COLLIDABLE_MIN = 5.0
CREATURE_CORPSE_DESPAWN_LIFECYCLE = -10.0


class CreatureLifecyclePhase(IntEnum):
    ALIVE = 0
    DEATH_STAGING = 1
    CORPSE_FADING = 2
    DESPAWNED = 3


def creature_lifecycle_is_alive(lifecycle_stage: float) -> bool:
    return float(lifecycle_stage) == CREATURE_LIFECYCLE_ALIVE


def creature_lifecycle_is_collidable(lifecycle_stage: float) -> bool:
    return float(lifecycle_stage) > CREATURE_LIFECYCLE_COLLIDABLE_MIN


def classify_creature_lifecycle(lifecycle_stage: float) -> CreatureLifecyclePhase:
    if creature_lifecycle_is_alive(float(lifecycle_stage)):
        return CreatureLifecyclePhase.ALIVE
    if float(lifecycle_stage) > 0.0:
        return CreatureLifecyclePhase.DEATH_STAGING
    if float(lifecycle_stage) >= CREATURE_CORPSE_DESPAWN_LIFECYCLE:
        return CreatureLifecyclePhase.CORPSE_FADING
    return CreatureLifecyclePhase.DESPAWNED


__all__ = [
    "CREATURE_CORPSE_DESPAWN_LIFECYCLE",
    "CREATURE_LIFECYCLE_ALIVE",
    "CREATURE_LIFECYCLE_COLLIDABLE_MIN",
    "CreatureLifecyclePhase",
    "classify_creature_lifecycle",
    "creature_lifecycle_is_alive",
    "creature_lifecycle_is_collidable",
]



from __future__ import annotations

import math
from collections.abc import Callable
from typing import TYPE_CHECKING

import msgspec

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rl
from grim.terrain_render import _maybe_alpha_test

from ...creatures.spawn import CreatureFlags, CreatureTypeId
from ...effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, SIZE_CODE_GRID, EffectId
from ...perks import PerkId
from ...perks.helpers import perk_active
from ...sim.world_defs import CREATURE_ANIM, CREATURE_ASSET
from ...ui.cursor import draw_aim_cursor
from .bonuses import draw_bonus_hover_labels, draw_bonus_pickups
from .constants import _RAD_TO_DEG, monster_vision_fade_alpha
from .context import WorldRenderCtx
from .creatures import draw_creature_sprite
from .effects import draw_effect_pool, draw_particle_pool, draw_sprite_effect_pool
from .overlays import draw_aim_circle, draw_clock_gauge, draw_direction_arrows
from .profile_hooks import profile_pass
from .projectiles import draw_projectile, draw_secondary_projectile, draw_sharpshooter_laser_sight
from .trooper import draw_player_trooper_sprite

if TYPE_CHECKING:
    from ...creatures.runtime import CreatureState
    from ...sim.state_types import PlayerState

class WorldDrawContext(msgspec.Struct, frozen=True):
    camera: Vec2 = Vec2()
    view_scale: Vec2 = Vec2(1.0, 1.0)
    scale: float = 1.0
    entity_alpha: float = 1.0
    trooper_texture: rl.Texture | None = None
    particles_texture: rl.Texture | None = None
    monster_vision: bool = False
    monster_vision_src: rl.Rectangle | None = None
    poison_src: rl.Rectangle | None = None


type _WorldToScreenWithFn = Callable[[Vec2, Vec2, Vec2], Vec2]
type _DrawAimCircleFn = Callable[[Vec2, float, float], None]
type _DrawClockGaugeFn = Callable[[Vec2, int, float, float], None]


def draw_world(
    render_ctx: WorldRenderCtx,
    *,
    draw_aim_indicators: bool = True,
    entity_alpha: float = 1.0,
) -> None:
    entity_alpha = clamp(float(entity_alpha), 0.0, 1.0)
    camera, view_scale, scale, screen_size, out_size = compute_view_transform(render_ctx)
    with profile_pass("background"):
        draw_background(render_ctx, camera=camera, screen_size=screen_size, out_size=out_size, view_scale=view_scale)
    if entity_alpha <= 1e-3:
        return

    with _maybe_alpha_test(bool(alpha_test_enabled(render_ctx))):
        draw_ctx = build_draw_context(
            render_ctx,
            camera=camera,
            view_scale=view_scale,
            scale=scale,
            entity_alpha=entity_alpha,
        )
        with profile_pass("players_dead"):
            draw_players(render_ctx, ctx=draw_ctx, alive=False)
        with profile_pass("creatures"):
            draw_creatures(render_ctx, ctx=draw_ctx)
        with profile_pass("freeze_overlay"):
            draw_freeze_overlay(render_ctx, ctx=draw_ctx)
        with profile_pass("players_alive"):
            draw_players(render_ctx, ctx=draw_ctx, alive=True)
        with profile_pass("projectiles_effects"):
            draw_projectiles_and_effects(render_ctx, ctx=draw_ctx)
        with profile_pass("bonus_ui"):
            draw_bonus_and_ui(render_ctx, ctx=draw_ctx, draw_aim_indicators_enabled=draw_aim_indicators)


def compute_view_transform(render_ctx: WorldRenderCtx) -> tuple[Vec2, Vec2, float, Vec2, Vec2]:
    out_w = float(rl.get_screen_width())
    out_h = float(rl.get_screen_height())
    out_size = Vec2(out_w, out_h)
    screen_size = render_ctx._camera_screen_size(runtime_w=out_w, runtime_h=out_h)
    camera = render_ctx._clamp_camera(render_ctx.camera, screen_size)
    scale_x = out_w / screen_size.x if screen_size.x > 0 else 1.0
    scale_y = out_h / screen_size.y if screen_size.y > 0 else 1.0
    view_scale = Vec2(scale_x, scale_y)
    scale = render_ctx._view_scale_avg(view_scale)
    return camera, view_scale, scale, screen_size, out_size


def draw_background(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    screen_size: Vec2,
    out_size: Vec2,
    view_scale: Vec2,
) -> None:
    clear_color = rl.Color(10, 10, 12, 255)
    ground = render_ctx.ground
    rl.clear_background(clear_color)
    if ground is not None:
        ground.draw(
            camera,
            screen_w=screen_size.x,
            screen_h=screen_size.y,
            out_w=out_size.x,
            out_h=out_size.y,
        )
        return

    # World bounds for debug if terrain is missing.
    world_min = camera.mul_components(view_scale)
    world_max = (camera + Vec2(float(render_ctx.world_size), float(render_ctx.world_size))).mul_components(view_scale)
    rl.draw_rectangle_lines(
        int(world_min.x),
        int(world_min.y),
        int(world_max.x - world_min.x),
        int(world_max.y - world_min.y),
        rl.Color(40, 40, 55, 255),
    )


def alpha_test_enabled(render_ctx: WorldRenderCtx) -> bool:
    ground = render_ctx.ground
    if ground is None:
        return True
    return ground.alpha_test


def effect_src_rect(texture: rl.Texture, effect_id: EffectId) -> rl.Rectangle | None:
    atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(effect_id))
    if atlas is None:
        return None
    grid = SIZE_CODE_GRID.get(int(atlas.size_code))
    if not grid:
        return None
    frame = int(atlas.frame)
    col = frame % grid
    row = frame // grid
    cell_w = float(texture.width) / float(grid)
    cell_h = float(texture.height) / float(grid)
    return rl.Rectangle(
        cell_w * float(col),
        cell_h * float(row),
        max(0.0, cell_w - 2.0),
        max(0.0, cell_h - 2.0),
    )


def build_draw_context(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    entity_alpha: float,
) -> WorldDrawContext:
    trooper_asset = CREATURE_ASSET.get(CreatureTypeId.TROOPER)
    trooper_texture = render_ctx.creature_textures.get(trooper_asset) if trooper_asset is not None else None
    particles_texture = render_ctx.particles_texture

    monster_vision = bool(render_ctx.players) and perk_active(render_ctx.players[0], PerkId.MONSTER_VISION)
    monster_vision_src = None
    if monster_vision and particles_texture is not None:
        monster_vision_src = effect_src_rect(particles_texture, EffectId.AURA)

    poison_src = None
    if particles_texture is not None:
        # Native uses `effect_select_texture(0x10)` (EffectId.AURA) for creature overlays
        # (monster vision, shadow, poison aura).
        poison_src = effect_src_rect(particles_texture, EffectId.AURA)

    return WorldDrawContext(
        camera=camera,
        view_scale=view_scale,
        scale=scale,
        entity_alpha=entity_alpha,
        trooper_texture=trooper_texture,
        particles_texture=particles_texture,
        monster_vision=monster_vision,
        monster_vision_src=monster_vision_src,
        poison_src=poison_src,
    )


def draw_player(render_ctx: WorldRenderCtx, player: PlayerState, *, ctx: WorldDrawContext) -> None:
    if ctx.trooper_texture is not None:
        draw_player_trooper_sprite(
            render_ctx,
            ctx.trooper_texture,
            player,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )
        return

    screen = render_ctx._world_to_screen_with(player.pos, camera=ctx.camera, view_scale=ctx.view_scale)
    tint = rl.Color(90, 190, 120, int(255 * ctx.entity_alpha + 0.5))
    rl.draw_circle(int(screen.x), int(screen.y), max(1.0, 14.0 * ctx.scale), tint)


def draw_players(render_ctx: WorldRenderCtx, *, ctx: WorldDrawContext, alive: bool) -> None:
    for player in render_ctx.players:
        if alive and player.health <= 0.0:
            continue
        if not alive and player.health > 0.0:
            continue
        draw_player(render_ctx, player, ctx=ctx)


def sorted_active_creatures(render_ctx: WorldRenderCtx) -> list[tuple[int, CreatureState]]:
    creature_type_order = {
        CreatureTypeId.ZOMBIE: 0,
        CreatureTypeId.SPIDER_SP1: 1,
        CreatureTypeId.SPIDER_SP2: 2,
        CreatureTypeId.ALIEN: 3,
        CreatureTypeId.LIZARD: 4,
        CreatureTypeId.TROOPER: 5,
    }
    creatures = [(idx, creature) for idx, creature in enumerate(render_ctx.creatures.entries) if creature.active]
    creatures.sort(
        key=lambda item: (creature_type_order[item[1].type_id], item[0]),
    )
    return creatures


def draw_creature_overlays(
    render_ctx: WorldRenderCtx,
    creature: CreatureState,
    *,
    screen: Vec2,
    lifecycle_stage: float,
    ctx: WorldDrawContext,
) -> None:
    fade = monster_vision_fade_alpha(lifecycle_stage)
    if ctx.monster_vision and ctx.particles_texture is not None and ctx.monster_vision_src is not None:
        mv_alpha = fade * ctx.entity_alpha
        if mv_alpha > 1e-3:
            size = 90.0 * ctx.scale
            dst = rl.Rectangle(screen.x, screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            tint = rl.Color(255, 255, 0, int(clamp(mv_alpha, 0.0, 1.0) * 255.0 + 0.5))
            rl.draw_texture_pro(ctx.particles_texture, ctx.monster_vision_src, dst, origin, 0.0, tint)

    if ctx.particles_texture is not None and ctx.poison_src is not None and bool(creature.plague_infected):
        # creature_render_all: collision_flag overlay (black 80x80 aura), drawn before red poison flag.
        plague_alpha = fade * ctx.entity_alpha
        if plague_alpha > 1e-3:
            size = 80.0 * ctx.scale
            dst = rl.Rectangle(screen.x, screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            tint = rl.Color(0, 0, 0, int(clamp(plague_alpha, 0.0, 1.0) * 255.0 + 0.5))
            rl.draw_texture_pro(ctx.particles_texture, ctx.poison_src, dst, origin, 0.0, tint)

    if (
        ctx.particles_texture is not None
        and ctx.poison_src is not None
        and (creature.flags & CreatureFlags.SELF_DAMAGE_TICK)
    ):
        poison_alpha = fade * ctx.entity_alpha
        if poison_alpha > 1e-3:
            size = 60.0 * ctx.scale
            dst = rl.Rectangle(screen.x, screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            tint = rl.Color(255, 0, 0, int(clamp(poison_alpha, 0.0, 1.0) * 255.0 + 0.5))
            rl.draw_texture_pro(ctx.particles_texture, ctx.poison_src, dst, origin, 0.0, tint)


def draw_creatures(render_ctx: WorldRenderCtx, *, ctx: WorldDrawContext) -> None:
    for _idx, creature in sorted_active_creatures(render_ctx):
        screen = render_ctx._world_to_screen_with(creature.pos, camera=ctx.camera, view_scale=ctx.view_scale)
        lifecycle_stage = float(creature.lifecycle_stage)

        type_id = creature.type_id
        asset = CREATURE_ASSET[type_id]
        texture = render_ctx.creature_textures.get(asset)

        draw_creature_overlays(render_ctx, creature, screen=screen, lifecycle_stage=lifecycle_stage, ctx=ctx)

        if texture is None:
            tint = rl.Color(220, 90, 90, int(255 * ctx.entity_alpha + 0.5))
            rl.draw_circle(int(screen.x), int(screen.y), max(1.0, creature.size * 0.5 * ctx.scale), tint)
            continue

        info = CREATURE_ANIM[type_id]

        tint_rgba = creature.tint

        # Energizer: tint "weak" creatures blue-ish while active.
        # Mirrors `creature_render_type` (0x00418b60) branch when
        # `_bonus_energizer_timer > 0` and `max_health < 500`.
        energizer_timer = float(render_ctx.state.bonuses.energizer)
        if energizer_timer > 0.0 and float(creature.max_hp) < 500.0:
            # Native clamps to 1.0, then blends towards (0.5, 0.5, 1.0, 1.0).
            # Effect is full strength while timer >= 1 and fades out during the last second.
            t = energizer_timer
            if t >= 1.0:
                t = 1.0
            elif t < 0.0:
                t = 0.0
            tint_rgba = RGBA.lerp(tint_rgba, RGBA(0.5, 0.5, 1.0, 1.0), t)

        if lifecycle_stage < 0.0:
            # Mirrors the main-pass alpha fade when lifecycle_stage ramps negative.
            tint_rgba = tint_rgba.with_alpha(max(0.0, tint_rgba.a + lifecycle_stage * 0.1))

        tint = tint_rgba.scaled_alpha(ctx.entity_alpha).clamped().to_rl()

        size_scale = clamp(float(creature.size) / 64.0, 0.25, 2.0)
        fx_detail = render_ctx.config.fx_detail(level=0, default=True) if render_ctx.config is not None else True
        # Mirrors `creature_render_type`: the "shadow-ish" pass is gated by fx_detail_0
        # and is disabled when the Monster Vision perk is active.
        shadow = fx_detail and (not render_ctx.players or not perk_active(render_ctx.players[0], PerkId.MONSTER_VISION))
        long_strip = (creature.flags & CreatureFlags.ANIM_PING_PONG) == 0 or (
            creature.flags & CreatureFlags.ANIM_LONG_STRIP
        ) != 0

        phase = float(creature.anim_phase)
        if long_strip:
            if lifecycle_stage < 0.0:
                # Negative phase selects the fallback "corpse" frame in creature_render_type.
                phase = -1.0
            elif lifecycle_stage < 16.0:
                # Death staging: while lifecycle_stage ramps down (16..0), creature_render_type
                # selects frames via `__ftol((base_frame + 15) - lifecycle_stage)`.
                phase = float(info.base + 0x0F) - lifecycle_stage - 0.5

        shadow_alpha = None
        if shadow:
            # Shadow pass uses tint_a * 0.4 and fades much faster for corpses (lifecycle_stage < 0).
            shadow_a = float(creature.tint.a) * 0.4
            if lifecycle_stage < 0.0:
                shadow_a += lifecycle_stage * (0.5 if long_strip else 0.1)
                shadow_a = max(0.0, shadow_a)
            shadow_alpha = int(clamp(shadow_a * ctx.entity_alpha * 255.0, 0.0, 255.0) + 0.5)

        draw_creature_sprite(
            render_ctx,
            texture,
            type_id=type_id or CreatureTypeId.ZOMBIE,
            flags=creature.flags,
            phase=phase,
            mirror_long=bool(info.mirror) and lifecycle_stage >= 16.0,
            shadow_alpha=shadow_alpha,
            pos=creature.pos,
            screen_pos=screen,
            rotation_rad=float(creature.heading) - math.pi / 2.0,
            scale=ctx.scale,
            size_scale=size_scale,
            tint=tint,
            shadow=shadow,
        )


def draw_freeze_overlay(render_ctx: WorldRenderCtx, *, ctx: WorldDrawContext) -> None:
    if ctx.particles_texture is None:
        return

    freeze_timer = float(render_ctx.state.bonuses.freeze)
    if freeze_timer <= 0.0:
        return

    src = effect_src_rect(ctx.particles_texture, EffectId.FREEZE_SHATTER)
    if src is None:
        return

    fade = 1.0 if freeze_timer >= 1.0 else clamp(freeze_timer, 0.0, 1.0)
    freeze_alpha = clamp(fade * ctx.entity_alpha * 0.7, 0.0, 1.0)
    if freeze_alpha <= 1e-3:
        return

    tint = rl.Color(255, 255, 255, int(freeze_alpha * 255.0 + 0.5))
    rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)
    for idx, creature in enumerate(render_ctx.creatures.entries):
        if not creature.active:
            continue
        size = float(creature.size) * ctx.scale
        if size <= 1e-3:
            continue
        creature_screen = render_ctx._world_to_screen_with(
            creature.pos,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
        )
        dst = rl.Rectangle(creature_screen.x, creature_screen.y, size, size)
        origin = rl.Vector2(size * 0.5, size * 0.5)
        rotation_deg = (float(idx) * 0.01 + float(creature.heading)) * _RAD_TO_DEG
        rl.draw_texture_pro(ctx.particles_texture, src, dst, origin, rotation_deg, tint)
    rl.end_blend_mode()


def draw_projectiles_and_effects(render_ctx: WorldRenderCtx, *, ctx: WorldDrawContext) -> None:
    with profile_pass("laser_sight"):
        draw_sharpshooter_laser_sight(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )

    with profile_pass("primary_projectiles"):
        for proj_index, proj in enumerate(render_ctx.state.projectiles.entries):
            if not proj.active:
                continue
            draw_projectile(
                render_ctx,
                proj,
                proj_index=proj_index,
                camera=ctx.camera,
                view_scale=ctx.view_scale,
                scale=ctx.scale,
                alpha=ctx.entity_alpha,
            )

    with profile_pass("particle_pool"):
        draw_particle_pool(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            alpha=ctx.entity_alpha,
        )

    with profile_pass("secondary_projectiles"):
        for proj in render_ctx.state.secondary_projectiles.entries:
            if not proj.active:
                continue
            draw_secondary_projectile(
                render_ctx,
                proj,
                camera=ctx.camera,
                view_scale=ctx.view_scale,
                scale=ctx.scale,
                alpha=ctx.entity_alpha,
            )

    with profile_pass("sprite_effect_pool"):
        draw_sprite_effect_pool(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            alpha=ctx.entity_alpha,
        )
    with profile_pass("effect_pool"):
        draw_effect_pool(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            alpha=ctx.entity_alpha,
        )


def iter_visible_aim_players(render_ctx: WorldRenderCtx) -> tuple[PlayerState, ...]:
    players = render_ctx.players
    if not bool(render_ctx.lan_local_aim_indicators_only):
        return tuple(players)
    local_slot = int(render_ctx.lan_local_player_slot_index)
    return tuple(player for player in players if int(player.index) == local_slot)


def draw_aim_indicators(
    render_ctx: WorldRenderCtx,
    *,
    ctx: WorldDrawContext,
    world_to_screen_with: _WorldToScreenWithFn | None = None,
    draw_aim_circle_fn: _DrawAimCircleFn | None = None,
    draw_clock_gauge_fn: _DrawClockGaugeFn | None = None,
) -> None:
    transform = world_to_screen_with
    if transform is None:
        def transform(pos: Vec2, camera: Vec2, view_scale: Vec2) -> Vec2:
            return render_ctx._world_to_screen_with(
                pos,
                camera=camera,
                view_scale=view_scale,
            )

    draw_circle = draw_aim_circle_fn
    if draw_circle is None:
        def draw_circle(center: Vec2, radius: float, alpha: float) -> None:
            draw_aim_circle(
                render_ctx,
                center=center,
                radius=radius,
                alpha=alpha,
            )

    draw_gauge = draw_clock_gauge_fn
    if draw_gauge is None:
        def draw_gauge(pos: Vec2, ms: int, scale: float, alpha: float) -> None:
            draw_clock_gauge(
                render_ctx,
                pos=pos,
                ms=ms,
                scale=scale,
                alpha=alpha,
            )

    for player in iter_visible_aim_players(render_ctx):
        if player.health <= 0.0:
            continue

        aim = player.aim
        dist = player.pos.distance_to(player.aim)
        radius = max(6.0, dist * float(player.spread_heat) * 0.5)
        aim_screen = transform(aim, ctx.camera, ctx.view_scale)
        screen_radius = max(1.0, radius * ctx.scale)
        draw_circle(aim_screen, screen_radius, ctx.entity_alpha)

        reload_timer = float(player.weapon.reload_timer)
        reload_max = float(player.weapon.reload_timer_max)
        if reload_max > 1e-6 and reload_timer > 1e-6:
            progress = reload_timer / reload_max
            if progress > 0.0:
                ms = int(progress * 60000.0)
                draw_gauge(Vec2(int(aim_screen.x), int(aim_screen.y)), ms, ctx.scale, ctx.entity_alpha)


def draw_aim_enhancements(
    render_ctx: WorldRenderCtx,
    *,
    ctx: WorldDrawContext,
    world_to_screen_with: _WorldToScreenWithFn | None = None,
) -> None:
    transform = world_to_screen_with
    if transform is None:
        def transform(pos: Vec2, camera: Vec2, view_scale: Vec2) -> Vec2:
            return render_ctx._world_to_screen_with(
                pos,
                camera=camera,
                view_scale=view_scale,
            )

    for player in iter_visible_aim_players(render_ctx):
        if player.health <= 0.0:
            continue
        aim_screen = transform(player.aim, ctx.camera, ctx.view_scale)
        draw_aim_cursor(ctx.particles_texture, render_ctx.aim_texture, pos=aim_screen)


def draw_bonus_and_ui(
    render_ctx: WorldRenderCtx,
    *,
    ctx: WorldDrawContext,
    draw_aim_indicators_enabled: bool,
) -> None:
    with profile_pass("bonus_pickups"):
        draw_bonus_pickups(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )
    with profile_pass("bonus_labels"):
        draw_bonus_hover_labels(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            alpha=ctx.entity_alpha,
        )

    draw_world_aim = draw_aim_indicators_enabled and (not render_ctx.demo_mode_active)
    if draw_world_aim:
        with profile_pass("aim_indicators"):
            draw_aim_indicators(render_ctx, ctx=ctx)

    with profile_pass("direction_arrows"):
        draw_direction_arrows(
            render_ctx,
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )

    if draw_world_aim:
        with profile_pass("aim_enhancements"):
            draw_aim_enhancements(render_ctx, ctx=ctx)


__all__ = [
    "WorldDrawContext",
    "draw_aim_enhancements",
    "draw_aim_indicators",
    "draw_world",
]



from __future__ import annotations

from grim.geom import Vec2

from ..sim.input import PlayerInput
from ..sim.state_types import GameplayState, PlayerState
from ..weapon_runtime import weapon_assign_player
from ..weapons import WeaponId

TYPO_WEAPON_ID = WeaponId.SAWED_OFF_SHOTGUN


def enforce_typo_player_frame(player: PlayerState, *, state: GameplayState) -> None:
    """Match Typ-o Shooter's bespoke player loop (`player_fire_weapon @ 0x00444980`).

    Typ-o resets timers and tops up ammo each frame, so typing speed (not weapon
    cooldown) controls rate of fire.
    """

    if player.weapon.weapon_id != TYPO_WEAPON_ID:
        weapon_assign_player(player, TYPO_WEAPON_ID, state=state)

    player.weapon.shot_cooldown = 0.0
    player.spread_heat = 0.0
    player.weapon.ammo = float(max(0, int(player.weapon.clip_size)))

    player.weapon.reload_active = False
    player.weapon.reload_timer = 0.0
    player.weapon.reload_timer_max = 0.0


def build_typo_player_input(
    *,
    aim: Vec2,
    fire_requested: bool,
    reload_requested: bool,
) -> PlayerInput:
    fire = bool(fire_requested)
    return PlayerInput(
        move=Vec2(),
        aim=aim,
        fire_down=fire,
        fire_pressed=fire,
        reload_pressed=bool(reload_requested),
    )



from __future__ import annotations

import math

from ..math_parity import f32


def _field_label(field: str | None) -> str:
    if field and str(field).strip():
        return str(field)
    return "value"


def wire_f32(value: float, *, field: str | None = None) -> float:
    label = _field_label(field)
    narrowed = float(f32(float(value)))
    if not math.isfinite(narrowed):
        raise ValueError(f"{label} must be finite")
    return narrowed


def wire_f32_opt(value: float | None, *, field: str | None = None) -> float | None:
    if value is None:
        return None
    return wire_f32(float(value), field=field)


def assert_wire_f32(value: float, *, field: str | None = None) -> float:
    label = _field_label(field)
    candidate = float(value)
    if not math.isfinite(candidate):
        raise ValueError(f"{label} must be finite")
    narrowed = float(f32(candidate))
    if candidate != narrowed:
        raise ValueError(f"{label} must be f32-canonical")
    return candidate



from __future__ import annotations

from typing import cast

import msgspec

from ..replay.checkpoints import ReplayCheckpoint
from .canonical_channels import (
    BonusEntitySample,
    CreatureEntitySample,
    EntitySamplesSnapshot,
    ProjectileEntitySample,
    RngStreamRow,
    SecondaryProjectileEntitySample,
    SimStateSnapshot,
    TimingSampleRow,
)
from .schema import TickRecord
from .trace import TraceError

ENTITY_SAMPLE_KINDS = ("creatures", "projectiles", "secondary_projectiles", "bonuses")

EntitySampleRow = CreatureEntitySample | ProjectileEntitySample | SecondaryProjectileEntitySample | BonusEntitySample

def _decode_channel(value: object, *, channel_name: str, target: object) -> object:
    try:
        return msgspec.convert(value, type=target)
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise TraceError(f"invalid {channel_name} channel payload") from exc


def _require_row(row: TickRecord | None, *, channel_name: str) -> TickRecord:
    if row is None:
        raise TraceError(f"missing tick row for {channel_name} channel")
    return row


def _require_channel_payload(row: TickRecord | None, *, channel_name: str) -> object:
    record = _require_row(row, channel_name=channel_name)
    if channel_name not in record.channels:
        raise TraceError(f"missing {channel_name} channel payload")
    return record.channels[channel_name]


def checkpoint_channel(row: TickRecord | None) -> ReplayCheckpoint | None:
    if row is None:
        return None
    payload = row.channels.get("checkpoint")
    if payload is None:
        return None
    return cast("ReplayCheckpoint", _decode_channel(payload, channel_name="checkpoint", target=ReplayCheckpoint))


def checkpoint_channel_required(row: TickRecord | None) -> ReplayCheckpoint:
    payload = _require_channel_payload(row, channel_name="checkpoint")
    return cast("ReplayCheckpoint", _decode_channel(payload, channel_name="checkpoint", target=ReplayCheckpoint))


def sim_state_channel(row: TickRecord | None) -> SimStateSnapshot | None:
    if row is None:
        return None
    payload = row.channels.get("sim_state")
    if payload is None:
        return None
    return cast("SimStateSnapshot", _decode_channel(payload, channel_name="sim_state", target=SimStateSnapshot))


def sim_state_channel_required(row: TickRecord | None) -> SimStateSnapshot:
    payload = _require_channel_payload(row, channel_name="sim_state")
    return cast("SimStateSnapshot", _decode_channel(payload, channel_name="sim_state", target=SimStateSnapshot))


def entity_samples_channel(row: TickRecord | None) -> EntitySamplesSnapshot | None:
    if row is None:
        return None
    payload = row.channels.get("entity_samples")
    if payload is None:
        return None
    return cast(
        "EntitySamplesSnapshot",
        _decode_channel(payload, channel_name="entity_samples", target=EntitySamplesSnapshot),
    )


def entity_samples_channel_required(row: TickRecord | None) -> EntitySamplesSnapshot:
    payload = _require_channel_payload(row, channel_name="entity_samples")
    return cast(
        "EntitySamplesSnapshot",
        _decode_channel(payload, channel_name="entity_samples", target=EntitySamplesSnapshot),
    )


def rng_stream_channel_required(row: TickRecord | None) -> list[RngStreamRow]:
    payload = _require_channel_payload(row, channel_name="rng_stream")
    return cast("list[RngStreamRow]", _decode_channel(payload, channel_name="rng_stream", target=list[RngStreamRow]))


def rng_marks_channel_required(row: TickRecord | None) -> dict[str, int]:
    payload = _require_channel_payload(row, channel_name="rng_marks")
    return cast("dict[str, int]", _decode_channel(payload, channel_name="rng_marks", target=dict[str, int]))


def timing_samples_channel_required(row: TickRecord | None) -> list[TimingSampleRow]:
    payload = _require_channel_payload(row, channel_name="timing_samples")
    return cast(
        "list[TimingSampleRow]",
        _decode_channel(payload, channel_name="timing_samples", target=list[TimingSampleRow]),
    )


def entity_rows(samples: EntitySamplesSnapshot, *, kind: str) -> list[EntitySampleRow]:
    match str(kind):
        case "creatures":
            return [*samples.creatures]
        case "projectiles":
            return [*samples.projectiles]
        case "secondary_projectiles":
            return [*samples.secondary_projectiles]
        case "bonuses":
            return [*samples.bonuses]
        case _:
            raise ValueError(f"unsupported entity sample kind: {kind!r}")



from __future__ import annotations

import re
import shutil
import subprocess
from functools import lru_cache
from pathlib import Path
from typing import Literal, TypeAlias

import msgspec

from ..aim_schemes import AimScheme, aim_scheme_from_value
from ..game_modes import GameMode
from ..math_parity import f32
from ..movement_controls import MovementControlType, movement_control_type_from_value
from ..sim.input_providers import GameCommand
from ..weapon_usage import WEAPON_USAGE_SLOT_COUNT
from ..weapons import WeaponId

REPLAY_FORMAT_VERSION = 9
ReplayFormatVersion: TypeAlias = Literal[9]

BootstrapKind: TypeAlias = Literal["none", "terrain_v1"]

WEAPON_USAGE_COUNT = WEAPON_USAGE_SLOT_COUNT

FIRE_DOWN_FLAG = 1 << 0
FIRE_PRESSED_FLAG = 1 << 1
RELOAD_PRESSED_FLAG = 1 << 2
RELOAD_DOWN_FLAG = 1 << 16
MOVE_KEYS_PRESENT_FLAG = 1 << 3
MOVE_FORWARD_FLAG = 1 << 4
MOVE_BACKWARD_FLAG = 1 << 5
TURN_LEFT_FLAG = 1 << 6
TURN_RIGHT_FLAG = 1 << 7
MOVE_MODE_PRESENT_FLAG = 1 << 8
MOVE_MODE_SHIFT = 9
MOVE_MODE_MASK = 0x7
AIM_SCHEME_PRESENT_FLAG = 1 << 12
AIM_SCHEME_SHIFT = 13
AIM_SCHEME_MASK = 0x7

InputQuantization: TypeAlias = Literal["f32"]

_RELEASE_VERSION_RE = re.compile(r"^\d+\.\d+\.\d+$")


def _head_points_at_release_tag(*, version: str, repo_root: Path, git_exe: str) -> bool:
    """Return True when HEAD is tagged as the release version."""

    if _RELEASE_VERSION_RE.fullmatch(str(version)) is None:
        return False

    tags_out = subprocess.check_output(
        [git_exe, "tag", "--points-at", "HEAD"],
        cwd=repo_root,
        stderr=subprocess.DEVNULL,
    )
    tags = {line.strip() for line in tags_out.decode("utf-8", errors="replace").splitlines() if line.strip()}
    return str(version) in tags or f"v{version}" in tags


@lru_cache(maxsize=1)
def current_replay_game_version() -> str:
    """Return replay header `game_version`.

    - Release-tagged HEAD: "<version>"
    - Git checkout non-release: "<version>+g<short_sha>"
    - No git metadata: "<version>"
    """

    from .. import __version__

    version = str(__version__)
    try:
        git_exe = shutil.which("git")
        if git_exe is None:
            return version
        repo_root = Path(__file__).resolve().parents[3]
        out = subprocess.check_output(
            [git_exe, "rev-parse", "--short=12", "HEAD"],
            cwd=repo_root,
            stderr=subprocess.DEVNULL,
        )
        build = out.decode("utf-8", errors="replace").strip()
        if not build:
            return version
        if _head_points_at_release_tag(version=version, repo_root=repo_root, git_exe=git_exe):
            return version
        if "+" in version:
            return f"{version}.g{build}"
        return f"{version}+g{build}"
    except (OSError, subprocess.CalledProcessError):
        return version


def _default_game_version() -> str:
    return current_replay_game_version()


def quantize_f32(value: float) -> float:
    return float(f32(float(value)))


def normalize_weapon_usage_counts(values: object) -> tuple[int, ...]:
    if not isinstance(values, (list, tuple)):
        return (0,) * WEAPON_USAGE_COUNT
    normalized: list[int] = [0] * WEAPON_USAGE_COUNT
    for idx, value in enumerate(values[:WEAPON_USAGE_COUNT]):
        try:
            normalized[idx] = int(value) & 0xFFFFFFFF
        except (TypeError, ValueError, OverflowError):
            normalized[idx] = 0
    return tuple(normalized)


def pack_input_flags(
    *,
    fire_down: bool,
    fire_pressed: bool,
    reload_pressed: bool,
    reload_down: bool = False,
    move_mode: MovementControlType | None = None,
    aim_scheme: AimScheme | None = None,
    move_forward_pressed: bool | None = None,
    move_backward_pressed: bool | None = None,
    turn_left_pressed: bool | None = None,
    turn_right_pressed: bool | None = None,
) -> int:
    flags = 0
    if fire_down:
        flags |= FIRE_DOWN_FLAG
    if fire_pressed:
        flags |= FIRE_PRESSED_FLAG
    if reload_pressed:
        flags |= RELOAD_PRESSED_FLAG
    if reload_down:
        flags |= RELOAD_DOWN_FLAG
    key_fields = (
        move_forward_pressed,
        move_backward_pressed,
        turn_left_pressed,
        turn_right_pressed,
    )
    if any(field is not None for field in key_fields):
        flags |= MOVE_KEYS_PRESENT_FLAG
        if bool(move_forward_pressed):
            flags |= MOVE_FORWARD_FLAG
        if bool(move_backward_pressed):
            flags |= MOVE_BACKWARD_FLAG
        if bool(turn_left_pressed):
            flags |= TURN_LEFT_FLAG
        if bool(turn_right_pressed):
            flags |= TURN_RIGHT_FLAG
    if move_mode is not None:
        flags |= MOVE_MODE_PRESENT_FLAG
        flags |= (int(move_mode) & MOVE_MODE_MASK) << MOVE_MODE_SHIFT
    if aim_scheme is not None:
        flags |= AIM_SCHEME_PRESENT_FLAG
        flags |= (int(aim_scheme) & AIM_SCHEME_MASK) << AIM_SCHEME_SHIFT
    return int(flags)


def unpack_input_flags(flags: int) -> tuple[bool, bool, bool, bool]:
    flags = int(flags)
    return (
        bool(flags & FIRE_DOWN_FLAG),
        bool(flags & FIRE_PRESSED_FLAG),
        bool(flags & RELOAD_PRESSED_FLAG),
        bool(flags & RELOAD_DOWN_FLAG),
    )


def unpack_input_move_key_flags(flags: int) -> tuple[bool | None, bool | None, bool | None, bool | None]:
    flags = int(flags)
    if not bool(flags & MOVE_KEYS_PRESENT_FLAG):
        return None, None, None, None
    return (
        bool(flags & MOVE_FORWARD_FLAG),
        bool(flags & MOVE_BACKWARD_FLAG),
        bool(flags & TURN_LEFT_FLAG),
        bool(flags & TURN_RIGHT_FLAG),
    )


def unpack_input_mode_flags(flags: int) -> tuple[MovementControlType | None, AimScheme | None]:
    flags = int(flags)
    move_mode: MovementControlType | None = None
    aim_scheme: AimScheme | None = None
    if bool(flags & MOVE_MODE_PRESENT_FLAG):
        move_mode = movement_control_type_from_value((flags >> MOVE_MODE_SHIFT) & MOVE_MODE_MASK)
    if bool(flags & AIM_SCHEME_PRESENT_FLAG):
        aim_scheme_raw = (flags >> AIM_SCHEME_SHIFT) & AIM_SCHEME_MASK
        if aim_scheme_raw == AIM_SCHEME_MASK:
            aim_scheme_raw = -1
        aim_scheme = aim_scheme_from_value(aim_scheme_raw)
    return move_mode, aim_scheme


PackedPlayerInput: TypeAlias = list[float | int]
PackedTickInputs: TypeAlias = list[PackedPlayerInput]


def unpack_packed_player_input(packed: PackedPlayerInput) -> tuple[float, float, float, float, int]:
    """Decode a compact replay input row into scalar values.

    Stored shape is `[move_x, move_y, aim_x, aim_y, flags]`.
    Returns `(move_x, move_y, aim_x, aim_y, flags)`.
    """

    def _require_num_f(value: object, *, field: str) -> float:
        if isinstance(value, bool):
            raise TypeError(f"{field} must be numeric, got bool")
        if isinstance(value, (int, float)):
            return float(value)
        raise TypeError(f"{field} must be numeric")

    def _require_num_i(value: object, *, field: str) -> int:
        if isinstance(value, bool):
            raise TypeError(f"{field} must be numeric, got bool")
        if isinstance(value, int):
            return int(value)
        if isinstance(value, float):
            return int(value)
        raise TypeError(f"{field} must be numeric")

    if len(packed) != 5:
        raise ValueError(f"packed replay input must have 5 fields, got {len(packed)}")

    mx = _require_num_f(packed[0], field="move_x")
    my = _require_num_f(packed[1], field="move_y")
    ax = _require_num_f(packed[2], field="aim_x")
    ay = _require_num_f(packed[3], field="aim_y")
    flags = _require_num_i(packed[4], field="flags")

    return mx, my, ax, ay, flags


class ReplayStatusSnapshot(msgspec.Struct, frozen=True):
    quest_unlock_index: int = 0
    quest_unlock_index_full: int = 0
    weapon_usage_counts: tuple[int, ...] = msgspec.field(default_factory=lambda: (0,) * WEAPON_USAGE_COUNT)


class ReplayClaimedStatsSnapshot(msgspec.Struct, frozen=True):
    complete: bool = False
    ticks: int = 0
    elapsed_ms: int = 0
    score_xp: int = 0
    kills: int = 0
    most_used_weapon_id: WeaponId = WeaponId.NONE
    shots_fired: int = 0
    shots_hit: int = 0


class ReplayHeader(msgspec.Struct, frozen=True):
    game_mode_id: GameMode
    seed: int
    replay_format_version: int = REPLAY_FORMAT_VERSION
    # Quests can recover their spawn script deterministically from the level id.
    # Leave empty for non-quest modes or legacy replays.
    quest_level: str = ""
    bootstrap_kind: BootstrapKind = "none"
    bootstrap_seed: int = 0
    game_version: str = msgspec.field(default_factory=_default_game_version)
    tick_rate: int = 60
    # Mirrors the native quest retry scaling counter (`quest_fail_retry_count`).
    quest_fail_retry_count: int = 0
    hardcore: bool = False
    preserve_bugs: bool = False
    detail_preset: int = 5
    gore_disabled: int = 0
    world_size: float = 1024.0
    player_count: int = 1
    status: ReplayStatusSnapshot = msgspec.field(default_factory=ReplayStatusSnapshot)
    claimed_stats: ReplayClaimedStatsSnapshot = msgspec.field(default_factory=ReplayClaimedStatsSnapshot)
    input_quantization: InputQuantization = "f32"


class ReplayTick(msgspec.Struct, frozen=True):
    dt: float
    inputs: PackedTickInputs
    commands: list[GameCommand] = []


class Replay(msgspec.Struct):
    header: ReplayHeader
    ticks: list[ReplayTick]



from __future__ import annotations

from collections.abc import Callable
from typing import Protocol

RenderDraw = Callable[[], None]
RenderPresent = Callable[[], None]


class RenderSink(Protocol):
    def open(self) -> None: ...

    def present(self) -> None: ...

    def flush(self) -> None: ...

    def close(self) -> None: ...


class WindowSink:
    """Default interactive sink; present/flush delegate to the frame loop."""

    def __init__(
        self,
        *,
        present_frame: RenderPresent | None = None,
    ) -> None:
        self._present_frame = present_frame
        self._opened = False

    def open(self) -> None:
        self._opened = True

    def present(self) -> None:
        if self._present_frame is None:
            return
        self._present_frame()

    def flush(self) -> None:
        pass

    def close(self) -> None:
        self._opened = False


class RenderPipeline:
    """Shared draw/present orchestration for live and replay contexts."""

    def __init__(
        self,
        *,
        sink: RenderSink,
        on_resize: Callable[[int, int], None] | None = None,
        begin_end_drawing: bool = False,
        begin_draw: Callable[[], None] | None = None,
        end_draw: Callable[[], None] | None = None,
    ) -> None:
        if bool(begin_end_drawing) and (begin_draw is None or end_draw is None):
            raise ValueError("begin_draw and end_draw are required when begin_end_drawing=True")
        self._sink = sink
        self._on_resize = on_resize
        self._begin_end_drawing = bool(begin_end_drawing)
        self._begin_draw = begin_draw
        self._end_draw = end_draw
        self._opened = False
        self._width = -1
        self._height = -1

    def open(self, *, width: int, height: int) -> None:
        if self._opened:
            return
        normalized_w = max(0, int(width))
        normalized_h = max(0, int(height))
        try:
            if self._on_resize is not None:
                self._on_resize(normalized_w, normalized_h)
            self._sink.open()
        except Exception:
            try:
                self._sink.close()
            except Exception:
                pass
            self._opened = False
            self._width = -1
            self._height = -1
            raise
        self._opened = True
        self._width = normalized_w
        self._height = normalized_h

    def _ensure_open(self, *, width: int, height: int) -> None:
        if self._opened:
            return
        self.open(width=int(width), height=int(height))

    def _resize_if_needed(self, *, width: int, height: int) -> None:
        normalized_w = max(0, int(width))
        normalized_h = max(0, int(height))
        if normalized_w == self._width and normalized_h == self._height:
            return
        if self._on_resize is not None:
            self._on_resize(normalized_w, normalized_h)
        self._width = normalized_w
        self._height = normalized_h

    def draw(self, *, draw_frame: RenderDraw, width: int, height: int) -> None:
        self._ensure_open(width=int(width), height=int(height))
        self._resize_if_needed(width=int(width), height=int(height))
        if self._begin_end_drawing:
            begin_draw = self._begin_draw
            end_draw = self._end_draw
            if begin_draw is None or end_draw is None:
                raise RuntimeError("render pipeline missing begin/end draw callbacks")
            begin_draw()
            try:
                draw_frame()
            finally:
                end_draw()
            return
        draw_frame()

    def present(self) -> None:
        if not self._opened:
            return
        self._sink.present()

    def render(self, *, draw_frame: RenderDraw, width: int, height: int) -> None:
        self.draw(draw_frame=draw_frame, width=int(width), height=int(height))
        self.present()

    def flush(self) -> None:
        if not self._opened:
            return
        self._sink.flush()

    def close(self) -> None:
        if not self._opened:
            return
        try:
            self._sink.close()
        finally:
            self._opened = False
            self._width = -1
            self._height = -1



from __future__ import annotations

from typing import cast

import msgspec

from ..replay.checkpoints import ReplayCheckpoint
from .trace import TraceError


def checkpoint_to_channel(checkpoint: ReplayCheckpoint) -> dict[str, object]:
    return cast("dict[str, object]", msgspec.to_builtins(checkpoint))


def channel_to_checkpoint(value: object) -> ReplayCheckpoint:
    try:
        return msgspec.convert(value, type=ReplayCheckpoint)
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise TraceError("invalid checkpoint channel payload") from exc



from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING

import msgspec

from ..types import Projectile, ProjectileTemplateId
from .behaviors import (
    _linger_default,
    _linger_gauss_gun,
    _linger_ion_cannon,
    _linger_ion_minigun,
    _linger_ion_rifle,
    _post_hit_ion_common,
    _post_hit_ion_rifle,
    _post_hit_plague_spreader,
    _post_hit_plasma_cannon,
    _post_hit_pulse_gun,
    _post_hit_shrinkifier,
    _pre_hit_splitter,
)

if TYPE_CHECKING:
    from .behaviors import _ProjectileHitInfo, _ProjectileUpdateCtx


class LingerRule(Enum):
    DEFAULT = auto()
    GAUSS = auto()
    ION_RIFLE = auto()
    ION_MINIGUN = auto()
    ION_CANNON = auto()


class PreHitRule(Enum):
    NONE = auto()
    SPLITTER = auto()


class PostHitRule(Enum):
    NONE = auto()
    ION_COMMON = auto()
    ION_RIFLE = auto()
    PLASMA_CANNON = auto()
    SHRINKIFIER = auto()
    PULSE_PUSH = auto()
    PLAGUE_INFECT = auto()


class PrimaryProjectileRule(msgspec.Struct, frozen=True):
    linger: LingerRule
    pre_hit: PreHitRule = PreHitRule.NONE
    post_hit: PostHitRule = PostHitRule.NONE
    stop_on_hit: bool = True
    emit_freeze_shard: bool = True
    reset_shock_chain_on_linger: bool = False


_DEFAULT_RULE = PrimaryProjectileRule(linger=LingerRule.DEFAULT)


PRIMARY_PROJECTILE_RULE_BY_TYPE_ID: dict[ProjectileTemplateId, PrimaryProjectileRule] = {
    ProjectileTemplateId.GAUSS_GUN: PrimaryProjectileRule(
        linger=LingerRule.GAUSS,
        stop_on_hit=False,
        emit_freeze_shard=False,
    ),
    ProjectileTemplateId.FIRE_BULLETS: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        stop_on_hit=False,
        emit_freeze_shard=False,
    ),
    ProjectileTemplateId.BLADE_GUN: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        stop_on_hit=False,
    ),
    ProjectileTemplateId.PULSE_GUN: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        post_hit=PostHitRule.PULSE_PUSH,
    ),
    ProjectileTemplateId.ION_RIFLE: PrimaryProjectileRule(
        linger=LingerRule.ION_RIFLE,
        post_hit=PostHitRule.ION_RIFLE,
        reset_shock_chain_on_linger=True,
    ),
    ProjectileTemplateId.ION_MINIGUN: PrimaryProjectileRule(
        linger=LingerRule.ION_MINIGUN,
        post_hit=PostHitRule.ION_COMMON,
        reset_shock_chain_on_linger=True,
    ),
    ProjectileTemplateId.ION_CANNON: PrimaryProjectileRule(
        linger=LingerRule.ION_CANNON,
        post_hit=PostHitRule.ION_COMMON,
    ),
    ProjectileTemplateId.SHRINKIFIER: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        post_hit=PostHitRule.SHRINKIFIER,
    ),
    ProjectileTemplateId.PLASMA_CANNON: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        post_hit=PostHitRule.PLASMA_CANNON,
    ),
    ProjectileTemplateId.SPLITTER_GUN: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        pre_hit=PreHitRule.SPLITTER,
    ),
    ProjectileTemplateId.PLAGUE_SPREADER: PrimaryProjectileRule(
        linger=LingerRule.DEFAULT,
        post_hit=PostHitRule.PLAGUE_INFECT,
    ),
}


def primary_rule_for_type_id(type_id: ProjectileTemplateId) -> PrimaryProjectileRule:
    return PRIMARY_PROJECTILE_RULE_BY_TYPE_ID.get(type_id, _DEFAULT_RULE)


def apply_linger(rule: LingerRule, ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    match rule:
        case LingerRule.DEFAULT:
            _linger_default(ctx, proj)
        case LingerRule.GAUSS:
            _linger_gauss_gun(ctx, proj)
        case LingerRule.ION_RIFLE:
            _linger_ion_rifle(ctx, proj)
        case LingerRule.ION_MINIGUN:
            _linger_ion_minigun(ctx, proj)
        case LingerRule.ION_CANNON:
            _linger_ion_cannon(ctx, proj)
        case _:
            raise AssertionError(f"Unsupported linger rule: {rule!r}")


def apply_pre_hit(rule: PreHitRule, ctx: _ProjectileUpdateCtx, proj: Projectile, hit_idx: int) -> None:
    match rule:
        case PreHitRule.NONE:
            return
        case PreHitRule.SPLITTER:
            _pre_hit_splitter(ctx, proj, hit_idx)
        case _:
            raise AssertionError(f"Unsupported pre-hit rule: {rule!r}")


def apply_post_hit(rule: PostHitRule, ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    match rule:
        case PostHitRule.NONE:
            return
        case PostHitRule.ION_COMMON:
            _post_hit_ion_common(ctx, hit)
        case PostHitRule.ION_RIFLE:
            _post_hit_ion_rifle(ctx, hit)
        case PostHitRule.PLASMA_CANNON:
            _post_hit_plasma_cannon(ctx, hit)
        case PostHitRule.SHRINKIFIER:
            _post_hit_shrinkifier(ctx, hit)
        case PostHitRule.PULSE_PUSH:
            _post_hit_pulse_gun(ctx, hit)
        case PostHitRule.PLAGUE_INFECT:
            _post_hit_plague_spreader(ctx, hit)
        case _:
            raise AssertionError(f"Unsupported post-hit rule: {rule!r}")


__all__ = [
    "PRIMARY_PROJECTILE_RULE_BY_TYPE_ID",
    "PrimaryProjectileRule",
    "apply_linger",
    "apply_post_hit",
    "apply_pre_hit",
    "primary_rule_for_type_id",
]



from __future__ import annotations

import shutil
from collections.abc import Callable
from pathlib import Path

import msgspec

from grim.raylib_api import rl

from .render_pipeline import RenderPipeline, WindowSink
from .view import View

SCREENSHOT_DIR = Path("screenshots")
SCREENSHOT_KEY = rl.KeyboardKey.KEY_F12


def _false_hook() -> bool:
    return False


class RunViewHooks(msgspec.Struct, frozen=True):
    should_close: Callable[[], bool] = _false_hook
    consume_screenshot_request: Callable[[], bool] = _false_hook


def _next_screenshot_index(directory: Path) -> int:
    if not directory.exists():
        return 1
    max_index = 0
    for entry in directory.glob("*.png"):
        stem = entry.stem
        if stem.isdigit():
            max_index = max(max_index, int(stem))
    return max_index + 1


def run_view(
    view: View,
    *,
    width: int = 1280,
    height: int = 720,
    title: str = "Crimsonland",
    fps: int = 60,
    config_flags: int = 0,
    exit_key: int | None = None,
    hooks: RunViewHooks | None = None,
) -> None:
    """Run a Raylib window with a pluggable debug view."""
    if config_flags:
        rl.set_config_flags(config_flags)
    rl.init_window(width, height, title)
    if exit_key is not None:
        rl.set_exit_key(exit_key)
    rl.set_target_fps(fps)
    run_hooks = hooks if hooks is not None else RunViewHooks()
    render_pipeline = RenderPipeline(
        sink=WindowSink(),
        begin_end_drawing=True,
        begin_draw=rl.begin_drawing,
        end_draw=rl.end_drawing,
    )
    try:
        view.open()
        screenshot_dir = SCREENSHOT_DIR if SCREENSHOT_DIR.is_absolute() else Path.cwd() / SCREENSHOT_DIR
        screenshot_index = _next_screenshot_index(screenshot_dir)
        while not rl.window_should_close():
            dt = rl.get_frame_time()
            view.update(dt)
            take_screenshot = rl.is_key_pressed(SCREENSHOT_KEY)
            if run_hooks.consume_screenshot_request():
                take_screenshot = True
            render_pipeline.draw(
                draw_frame=view.draw,
                width=rl.get_render_width(),
                height=rl.get_render_height(),
            )
            render_pipeline.present()
            if run_hooks.should_close():
                break
            if take_screenshot:
                screenshot_dir.mkdir(parents=True, exist_ok=True)
                filename = f"{screenshot_index:05d}.png"
                rl.take_screenshot(filename)
                src = Path.cwd() / filename
                if src.exists():
                    shutil.move(str(src), str(screenshot_dir / filename))
                screenshot_index += 1
    finally:
        try:
            view.close()
        finally:
            render_pipeline.close()
            rl.close_window()


def run_window(
    width: int = 1280,
    height: int = 720,
    title: str = "Crimsonland",
    fps: int = 60,
) -> None:
    """Open a minimal Raylib window for the reference implementation."""

    class _EmptyView:
        def open(self) -> None:
            return None

        def update(self, dt: float) -> None:
            return None

        def draw(self) -> None:
            rl.clear_background(rl.BLACK)

        def close(self) -> None:
            return None

    run_view(_EmptyView(), width=width, height=height, title=title, fps=fps)



from __future__ import annotations

from grim.geom import Vec2

from ...projectiles.types import Projectile

RAD_TO_DEG = 57.29577951308232


def proj_origin(proj: Projectile, fallback: Vec2) -> Vec2:
    _ = fallback
    return proj.origin


__all__ = ["RAD_TO_DEG", "proj_origin"]



from __future__ import annotations

import hashlib
import math
from contextlib import suppress
from datetime import UTC, datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import BinaryIO, cast

import msgspec

from grim.rand import CrtRand

from ..game_modes import GameMode
from ..net.session_settings import session_settings_for_lockstep
from ..replay.checkpoints import ReplayCheckpoint
from ..replay.codec import dump_replay_file
from ..replay.header_settings import replay_header_from_session_settings
from ..replay.types import WEAPON_USAGE_COUNT, BootstrapKind, Replay, ReplayStatusSnapshot, ReplayTick
from ..sim.bootstrap import run_terrain_bootstrap
from ..status_snapshot import progress_status_from_debug_snapshot, replay_status_from_progress
from .canonical_channels import EntitySamplesSnapshot, RngStreamRow, SimStateSnapshot, TimingSampleRow
from .rng import canonical_rng_marks
from .schema import (
    TRACE_FORMAT_VERSION,
    TRACE_SCHEMA_VERSION,
    TickRecord,
    TraceMeta,
    channel_versions_for,
)
from .trace import TraceSummary, write_trace_iter

_FRAME_LEN_BYTES = 4
_TICK_ENCODER = msgspec.msgpack.Encoder()
_TICK_DECODER = msgspec.msgpack.Decoder(type=TickRecord)
_GAME_MODE_QUESTS = 3
_GAME_MODE_SURVIVAL = int(GameMode.SURVIVAL)
_GAME_MODE_RUSH = int(GameMode.RUSH)
_TERRAIN_BOOTSTRAP_MODES = {_GAME_MODE_SURVIVAL, _GAME_MODE_RUSH}
_BOOTSTRAP_KINDS: set[BootstrapKind] = {"none", "terrain_v1"}
_SUPPORTED_CAPTURE_FORMAT_VERSION = 9
_MODE_LABEL_BY_ID = {
    int(GameMode.DEMO): "demo",
    int(GameMode.SURVIVAL): "survival",
    int(GameMode.RUSH): "rush",
    int(GameMode.QUESTS): "quests",
    int(GameMode.TYPO): "typo",
    int(GameMode.TUTORIAL): "tutorial",
}


class FridaFinalizeError(ValueError):
    pass


class _TickChannels(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    checkpoint: ReplayCheckpoint
    sim_state: SimStateSnapshot
    entity_samples: EntitySamplesSnapshot
    rng_marks: dict[str, int] = msgspec.field(default_factory=dict)
    rng_stream: list[RngStreamRow] = msgspec.field(default_factory=list)
    timing_samples: list[TimingSampleRow] = msgspec.field(default_factory=list)


class _SessionStartRow(
    msgspec.Struct,
    frozen=True,
    forbid_unknown_fields=True,
    tag_field="event",
    tag="session_start",
):
    capture_format_version: int
    schema_version: int = 1
    session_id: str = ""
    out_path: str = ""
    platform: str = "windows"
    arch: str = "x86"
    script_version: str = ""
    config: dict[str, object] = msgspec.field(default_factory=dict)
    session_fingerprint: dict[str, object] = msgspec.field(default_factory=dict)


class _RunStartRow(
    msgspec.Struct,
    frozen=True,
    forbid_unknown_fields=True,
    tag_field="event",
    tag="run_start",
):
    run_id: int
    mode_id: int
    seed: int
    player_count: int
    reason: str = "run_start"
    quest_stage_major: int = -1
    quest_stage_minor: int = -1
    bootstrap_kind: str = "none"
    bootstrap_seed: int = 0
    seed_source: str = "unknown"
    tick_index_global: int | None = None


class _TickRow(
    msgspec.Struct,
    frozen=True,
    forbid_unknown_fields=True,
    tag_field="event",
    tag="tick",
):
    run_id: int
    elapsed_ms: int
    dt: float
    dt_ms_i32: int
    mode_id: int
    channels: _TickChannels
    tick_index_global: int | None = None
    quest_stage_major: int = -1
    quest_stage_minor: int = -1
    phase_markers: list[str] = msgspec.field(default_factory=list)
    replay_inputs: list[tuple[float, float, float, float, int]] = msgspec.field(default_factory=list)


class _RunEndRow(
    msgspec.Struct,
    frozen=True,
    forbid_unknown_fields=True,
    tag_field="event",
    tag="run_end",
):
    run_id: int
    reason: str = "run_end"
    mode_id: int = -1
    quest_stage_major: int = -1
    quest_stage_minor: int = -1
    tick_index_global: int | None = None
    ticks_written: int = 0


class _SessionEndRow(
    msgspec.Struct,
    frozen=True,
    forbid_unknown_fields=True,
    tag_field="event",
    tag="session_end",
):
    session_id: str = ""
    ticks_written: int = 0


type _CaptureRow = _SessionStartRow | _RunStartRow | _TickRow | _RunEndRow | _SessionEndRow


_CAPTURE_ROW_DECODER = msgspec.json.Decoder(type=_CaptureRow)


class FinalizedTrace(msgspec.Struct, frozen=True):
    run_id: int
    out_path: Path
    replay_path: Path
    tick_count: int
    mode_id: int
    quest_stage_major: int
    quest_stage_minor: int
    summary: TraceSummary


class FinalizeResult(msgspec.Struct, frozen=True):
    raw_path: Path
    traces: list[FinalizedTrace]
    deleted_raw: bool


class _OpenRun(msgspec.Struct):
    run_id: int
    mode_id: int
    quest_stage_major: int
    quest_stage_minor: int
    replay_seed: int
    replay_bootstrap_kind: BootstrapKind
    replay_bootstrap_seed: int
    replay_player_count: int
    temp_path: Path
    stream: BinaryIO
    replay_seed_source: str = "unknown"
    tick_count: int = 0
    next_local_tick: int = 0
    replay_inputs: list[list[list[float | int]]] = msgspec.field(default_factory=list)
    replay_dt: list[float] = msgspec.field(default_factory=list)
    replay_status: ReplayStatusSnapshot = msgspec.field(default_factory=lambda: ReplayStatusSnapshot())
    channels_seen: set[str] = msgspec.field(default_factory=set)
    global_tick_first: int | None = None
    global_tick_last: int | None = None


def _fingerprint(path: Path) -> dict[str, object]:
    stat = path.stat()
    raw = path.read_bytes()
    return {
        "path": str(path),
        "sha256": hashlib.sha256(raw).hexdigest(),
        "size": int(stat.st_size),
        "mtime_ns": int(stat.st_mtime_ns),
    }


def _decode_capture_row(line: bytes, *, field: str) -> _CaptureRow:
    try:
        return _CAPTURE_ROW_DECODER.decode(line)
    except (msgspec.DecodeError, msgspec.ValidationError) as exc:
        raise FridaFinalizeError(f"{field} invalid capture row: {exc}") from exc


def _validate_bootstrap_kind(kind: str, *, field: str) -> BootstrapKind:
    if kind not in _BOOTSTRAP_KINDS:
        raise FridaFinalizeError(f"{field} must be one of {sorted(_BOOTSTRAP_KINDS)!r}, got {kind!r}")
    return cast("BootstrapKind", kind)


def _canonical_channels_payload(
    *,
    channels: _TickChannels,
    local_tick: int,
    field: str,
) -> tuple[ReplayCheckpoint, dict[str, object]]:
    checkpoint = msgspec.structs.replace(
        channels.checkpoint,
        tick_index=int(local_tick),
    )
    expected_rng_marks = canonical_rng_marks(
        rng_state=int(checkpoint.rng_state),
        rng_stream=channels.rng_stream,
    )
    if dict(checkpoint.rng_marks) != expected_rng_marks:
        raise FridaFinalizeError(
            f"{field}.checkpoint.rng_marks must match canonical rng marks; "
            "recapture with updated gameplay_diff_capture.js",
        )
    if dict(channels.rng_marks) != expected_rng_marks:
        raise FridaFinalizeError(
            f"{field}.rng_marks must match canonical rng marks; "
            "recapture with updated gameplay_diff_capture.js",
        )
    checkpoint = msgspec.structs.replace(checkpoint, rng_marks=dict(expected_rng_marks))
    normalized = _TickChannels(
        checkpoint=checkpoint,
        sim_state=channels.sim_state,
        entity_samples=channels.entity_samples,
        rng_marks=dict(expected_rng_marks),
        rng_stream=list(channels.rng_stream),
        timing_samples=list(channels.timing_samples),
    )
    built = msgspec.to_builtins(normalized)
    return checkpoint, cast("dict[str, object]", built)


def _replay_tick_inputs_from_row(
    replay_inputs: list[tuple[float, float, float, float, int]],
    *,
    expected_players: int,
    field: str,
) -> list[list[float | int]]:
    if len(replay_inputs) != int(expected_players):
        raise FridaFinalizeError(
            f"{field} must contain {int(expected_players)} player rows, got {len(replay_inputs)}",
        )
    out: list[list[float | int]] = []
    for move_x, move_y, aim_x, aim_y, flags in replay_inputs:
        out.append([float(move_x), float(move_y), float(aim_x), float(aim_y), int(flags)])
    return out


def _replay_status_from_channels(channels: _TickChannels) -> ReplayStatusSnapshot:
    status = channels.sim_state.gameplay.status
    expected_usage_count = int(WEAPON_USAGE_COUNT)
    if len(status.weapon_usage_counts) != expected_usage_count:
        raise FridaFinalizeError(
            "sim_state.gameplay.status.weapon_usage_counts length mismatch: "
            f"expected {expected_usage_count}, got {len(status.weapon_usage_counts)}",
        )
    progress = progress_status_from_debug_snapshot(status)
    return replay_status_from_progress(progress)


def _tick_iter_from_spool(path: Path):
    with Path(path).open("rb") as handle:
        while True:
            len_raw = handle.read(_FRAME_LEN_BYTES)
            if not len_raw:
                break
            if len(len_raw) != _FRAME_LEN_BYTES:
                raise FridaFinalizeError(f"truncated tick spool frame length: {path}")
            frame_len = int.from_bytes(len_raw, "little", signed=False)
            if frame_len <= 0:
                raise FridaFinalizeError(f"invalid tick spool frame length: {frame_len}")
            payload = handle.read(frame_len)
            if len(payload) != frame_len:
                raise FridaFinalizeError(f"truncated tick spool payload: {path}")
            try:
                yield _TICK_DECODER.decode(payload)
            except (msgspec.DecodeError, msgspec.ValidationError) as exc:
                raise FridaFinalizeError(f"invalid tick spool payload in {path}") from exc


def _run_output_path(
    *,
    raw_path: Path,
    output_dir: Path,
    mode_id: int,
    quest_stage_major: int,
    quest_stage_minor: int,
    counters: dict[str, int],
) -> Path:
    base = Path(raw_path).stem
    is_quest_run = (
        int(mode_id) == int(_GAME_MODE_QUESTS)
        and int(quest_stage_major) > 0
        and int(quest_stage_minor) > 0
    )
    if is_quest_run:
        key = f"quest_{int(quest_stage_major)}_{int(quest_stage_minor)}"
        idx = counters.get(key, 0) + 1
        counters[key] = idx
        name = f"{base}.quest_{int(quest_stage_major)}_{int(quest_stage_minor)}.run{idx}.cdt"
    else:
        mode_label = str(_MODE_LABEL_BY_ID.get(int(mode_id), f"mode_{int(mode_id)}"))
        key = mode_label
        idx = counters.get(key, 0) + 1
        counters[key] = idx
        name = f"{base}.{mode_label}.run{idx}.cdt"
    return Path(output_dir) / name


def _validate_run_seed_for_replay(run: _OpenRun) -> None:
    if int(run.mode_id) not in _TERRAIN_BOOTSTRAP_MODES:
        return
    bootstrap_seed = int(run.replay_bootstrap_seed)
    world_size = 1024
    quest_unlock_index = int(run.replay_status.quest_unlock_index)
    rng = CrtRand(seed=int(bootstrap_seed))
    terrain = run_terrain_bootstrap(
        rng,
        quest_unlock_index=int(quest_unlock_index),
        width=int(world_size),
        height=int(world_size),
        layers=3,
    )
    if int(run.replay_seed) != int(terrain.seed_after):
        raise FridaFinalizeError(
            f"run {run.run_id}: terrain bootstrap seed mismatch "
            f"mode_id={run.mode_id} "
            f"bootstrap_seed={bootstrap_seed} "
            f"quest_unlock_index={quest_unlock_index} "
            f"expected_seed_after={int(terrain.seed_after)} "
            f"run_start.seed={int(run.replay_seed)} "
            f"seed_source={run.replay_seed_source!r}",
        )


def _build_meta(
    *,
    raw_fingerprint: dict[str, object],
    session_start: _SessionStartRow,
    run: _OpenRun,
    tick_count: int,
    channels_seen: set[str],
) -> TraceMeta:
    producer_platform = str(session_start.platform)
    producer_arch = str(session_start.arch)
    producer_impl_version = str(session_start.script_version)
    config = dict(session_start.config)
    source = dict(raw_fingerprint)
    source["run_id"] = int(run.run_id)
    source["mode_id"] = int(run.mode_id)
    source["quest_stage_major"] = int(run.quest_stage_major)
    source["quest_stage_minor"] = int(run.quest_stage_minor)
    source["global_tick_first"] = None if run.global_tick_first is None else int(run.global_tick_first)
    source["global_tick_last"] = None if run.global_tick_last is None else int(run.global_tick_last)
    source["run_start_seed_source"] = str(run.replay_seed_source)

    sorted_channels = sorted(str(channel) for channel in channels_seen)
    tick_end = int(tick_count) - 1
    return TraceMeta(
        trace_format_version=int(TRACE_FORMAT_VERSION),
        trace_schema_version=int(TRACE_SCHEMA_VERSION),
        created_utc=datetime.now(tz=UTC).isoformat(),
        producer={
            "impl": "frida_original",
            "impl_version": producer_impl_version,
            "platform": producer_platform,
            "arch": producer_arch,
        },
        source=source,
        channels=sorted_channels,
        channel_versions=channel_versions_for(sorted_channels),
        tick_range={
            "start_tick": 0 if tick_count > 0 else -1,
            "end_tick": tick_end if tick_count > 0 else -1,
            "tick_count": int(tick_count),
        },
        config=config,
    )


def _write_run_trace(
    *,
    raw_path: Path,
    output_dir: Path,
    raw_fingerprint: dict[str, object],
    session_start: _SessionStartRow,
    run: _OpenRun,
    chunk_ticks: int,
    counters: dict[str, int],
) -> FinalizedTrace:
    run.stream.flush()
    run.stream.close()
    if int(run.replay_player_count) <= 0:
        raise FridaFinalizeError(f"run {run.run_id}: invalid replay player_count={run.replay_player_count}")
    if len(run.replay_inputs) != int(run.tick_count):
        raise FridaFinalizeError(
            f"run {run.run_id}: replay_inputs count {len(run.replay_inputs)} does not match tick_count {run.tick_count}",
        )
    if len(run.replay_dt) != int(run.tick_count):
        raise FridaFinalizeError(
            f"run {run.run_id}: replay_dt count {len(run.replay_dt)} "
            f"does not match tick_count {run.tick_count}",
        )
    _validate_run_seed_for_replay(run)
    out_path = _run_output_path(
        raw_path=raw_path,
        output_dir=output_dir,
        mode_id=int(run.mode_id),
        quest_stage_major=int(run.quest_stage_major),
        quest_stage_minor=int(run.quest_stage_minor),
        counters=counters,
    )
    meta = _build_meta(
        raw_fingerprint=raw_fingerprint,
        session_start=session_start,
        run=run,
        tick_count=int(run.tick_count),
        channels_seen=set(run.channels_seen),
    )
    summary = write_trace_iter(
        out_path,
        meta=meta,
        ticks=_tick_iter_from_spool(run.temp_path),
        chunk_ticks=max(1, int(chunk_ticks)),
    )
    replay_path = Path(out_path).with_suffix(".crd")
    is_quest_run = (
        int(run.mode_id) == int(_GAME_MODE_QUESTS)
        and int(run.quest_stage_major) > 0
        and int(run.quest_stage_minor) > 0
    )
    settings = session_settings_for_lockstep(
        mode_id=int(run.mode_id),
        player_count=int(run.replay_player_count),
        quest_level=(f"{int(run.quest_stage_major)}.{int(run.quest_stage_minor)}" if is_quest_run else ""),
        preserve_bugs=False,
    )
    replay_header = replay_header_from_session_settings(
        settings,
        seed=int(run.replay_seed),
        bootstrap_kind=run.replay_bootstrap_kind,
        bootstrap_seed=int(run.replay_bootstrap_seed),
        status=run.replay_status,
    )
    replay_ticks = [
        ReplayTick(
            inputs=run.replay_inputs[i],
            dt=run.replay_dt[i],
        )
        for i in range(run.tick_count)
    ]
    dump_replay_file(
        replay_path,
        Replay(header=replay_header, ticks=replay_ticks),
    )
    return FinalizedTrace(
        run_id=int(run.run_id),
        out_path=Path(out_path),
        replay_path=Path(replay_path),
        tick_count=int(run.tick_count),
        mode_id=int(run.mode_id),
        quest_stage_major=int(run.quest_stage_major),
        quest_stage_minor=int(run.quest_stage_minor),
        summary=summary,
    )


def finalize_frida_jsonl_to_traces(
    raw_path: Path,
    *,
    output_dir: Path | None = None,
    chunk_ticks: int = 256,
    delete_raw: bool = True,
) -> FinalizeResult:
    raw_path = Path(raw_path)
    if not raw_path.is_file():
        raise FridaFinalizeError(f"raw frida trace not found: {raw_path}")
    output_root = raw_path.parent if output_dir is None else Path(output_dir)
    output_root.mkdir(parents=True, exist_ok=True)
    raw_fingerprint = _fingerprint(raw_path)

    traces: list[FinalizedTrace] = []
    run_counters: dict[str, int] = {}
    session_start: _SessionStartRow | None = None
    session_ended = False
    active_run: _OpenRun | None = None

    temp_dir_obj = TemporaryDirectory(prefix="crimson-frida-finalize-")
    temp_root = Path(temp_dir_obj.name)
    try:
        with raw_path.open("rb") as handle:
            for line_no, raw_line in enumerate(handle, start=1):
                line = bytes(raw_line).strip()
                if not line:
                    continue
                row = _decode_capture_row(line, field=f"{raw_path}.lines[{line_no}]")

                match row:
                    case _SessionStartRow() as session_row:
                        if session_start is not None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] duplicate session_start")
                        if int(session_row.capture_format_version) != int(_SUPPORTED_CAPTURE_FORMAT_VERSION):
                            raise FridaFinalizeError(
                                f"{raw_path}.lines[{line_no}] unsupported capture_format_version="
                                f"{int(session_row.capture_format_version)}; expected {int(_SUPPORTED_CAPTURE_FORMAT_VERSION)}",
                            )
                        session_start = session_row
                        continue
                    case _:
                        if session_start is None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] must start with session_start")

                match row:
                    case _RunStartRow() as run_start:
                        if active_run is not None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] run_start while run is active")
                        bootstrap_kind = _validate_bootstrap_kind(
                            run_start.bootstrap_kind,
                            field=f"{raw_path}.lines[{line_no}].bootstrap_kind",
                        )
                        mode_id = int(run_start.mode_id)
                        if mode_id in _TERRAIN_BOOTSTRAP_MODES:
                            if bootstrap_kind != "terrain_v1":
                                raise FridaFinalizeError(
                                    f"{raw_path}.lines[{line_no}] mode_id={mode_id} requires "
                                    "bootstrap_kind='terrain_v1'; recapture with updated gameplay_diff_capture.js",
                                )
                        elif bootstrap_kind != "none":
                            raise FridaFinalizeError(
                                f"{raw_path}.lines[{line_no}] mode_id={mode_id} requires bootstrap_kind='none'",
                            )
                        if int(run_start.bootstrap_seed) < 0:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}].bootstrap_seed must be >= 0")
                        if int(run_start.player_count) <= 0:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}].player_count must be positive")
                        spool_path = temp_root / f"run_{int(run_start.run_id)}.ticks"
                        active_run = _OpenRun(
                            run_id=int(run_start.run_id),
                            mode_id=mode_id,
                            quest_stage_major=int(run_start.quest_stage_major),
                            quest_stage_minor=int(run_start.quest_stage_minor),
                            replay_seed=int(run_start.seed),
                            replay_bootstrap_kind=bootstrap_kind,
                            replay_bootstrap_seed=int(run_start.bootstrap_seed),
                            replay_player_count=int(run_start.player_count),
                            temp_path=spool_path,
                            stream=spool_path.open("wb"),
                            replay_seed_source=str(run_start.seed_source),
                        )
                        continue
                    case _TickRow() as tick_row:
                        if active_run is None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] tick outside run")
                        if int(tick_row.run_id) != int(active_run.run_id):
                            raise FridaFinalizeError(
                                f"{raw_path}.lines[{line_no}] tick run_id={int(tick_row.run_id)} "
                                f"does not match active run {active_run.run_id}",
                            )
                        if not math.isfinite(float(tick_row.dt)) or float(tick_row.dt) < 0.0:
                            raise FridaFinalizeError(
                                f"{raw_path}.lines[{line_no}].dt must be finite and >= 0",
                            )
                        if int(tick_row.dt_ms_i32) < 0:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}].dt_ms_i32 must be >= 0")
                        replay_inputs = _replay_tick_inputs_from_row(
                            tick_row.replay_inputs,
                            expected_players=int(active_run.replay_player_count),
                            field=f"{raw_path}.lines[{line_no}].replay_inputs",
                        )
                        _checkpoint, channels = _canonical_channels_payload(
                            channels=tick_row.channels,
                            local_tick=int(active_run.next_local_tick),
                            field=f"{raw_path}.lines[{line_no}].channels",
                        )
                        active_run.replay_status = _replay_status_from_channels(tick_row.channels)
                        active_run.replay_inputs.append(list(replay_inputs))
                        active_run.replay_dt.append(float(tick_row.dt))
                        tick = TickRecord(
                            tick_index=int(active_run.next_local_tick),
                            elapsed_ms=int(tick_row.elapsed_ms),
                            dt_ms_i32=int(tick_row.dt_ms_i32),
                            mode_id=int(tick_row.mode_id),
                            phase_markers=list(tick_row.phase_markers),
                            channels=channels,
                        )
                        payload = _TICK_ENCODER.encode(tick)
                        active_run.stream.write(len(payload).to_bytes(_FRAME_LEN_BYTES, "little", signed=False))
                        active_run.stream.write(payload)
                        active_run.next_local_tick += 1
                        active_run.tick_count += 1
                        active_run.channels_seen.update(str(name) for name in channels.keys())
                        if tick_row.tick_index_global is not None:
                            global_tick = int(tick_row.tick_index_global)
                            if active_run.global_tick_first is None:
                                active_run.global_tick_first = global_tick
                            active_run.global_tick_last = global_tick
                        continue
                    case _RunEndRow() as run_end:
                        if active_run is None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] run_end without active run")
                        if int(run_end.run_id) != int(active_run.run_id):
                            raise FridaFinalizeError(
                                f"{raw_path}.lines[{line_no}] run_end run_id={int(run_end.run_id)} "
                                f"does not match active run {active_run.run_id}",
                            )
                        traces.append(
                            _write_run_trace(
                                raw_path=raw_path,
                                output_dir=output_root,
                                raw_fingerprint=raw_fingerprint,
                                session_start=session_start,
                                run=active_run,
                                chunk_ticks=max(1, int(chunk_ticks)),
                                counters=run_counters,
                            ),
                        )
                        active_run = None
                        continue
                    case _SessionEndRow():
                        if active_run is not None:
                            raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] session_end while run is active")
                        session_ended = True
                        continue
                    case _:
                        raise FridaFinalizeError(f"{raw_path}.lines[{line_no}] unsupported capture row")

        if session_start is None:
            raise FridaFinalizeError(f"{raw_path} missing session_start")
        if active_run is not None:
            # Allow abrupt host/process shutdown to still produce a usable run.
            # We already validated all parsed rows and can finalize the in-flight spool.
            traces.append(
                _write_run_trace(
                    raw_path=raw_path,
                    output_dir=output_root,
                    raw_fingerprint=raw_fingerprint,
                    session_start=session_start,
                    run=active_run,
                    chunk_ticks=max(1, int(chunk_ticks)),
                    counters=run_counters,
                ),
            )
            active_run = None
        if not session_ended and not traces:
            raise FridaFinalizeError(f"{raw_path} missing session_end")
        if not traces:
            raise FridaFinalizeError(f"{raw_path} had no finalized runs")
    finally:
        if active_run is not None:
            with suppress(Exception):
                active_run.stream.close()
        with suppress(OSError):
            temp_dir_obj.cleanup()

    deleted = False
    if bool(delete_raw):
        raw_path.unlink(missing_ok=False)
        deleted = True

    return FinalizeResult(
        raw_path=raw_path,
        traces=list(traces),
        deleted_raw=bool(deleted),
    )



from __future__ import annotations

from typing import cast

import msgspec

from grim.raylib_api import rl

from ..terrain_assets import TerrainTextureId, terrain_texture_by_id
from .render_resources import RenderResources

DEFAULT_TERRAIN_IDS = (
    TerrainTextureId.Q1_BASE,
    TerrainTextureId.Q1_OVERLAY,
    TerrainTextureId.Q1_BASE,
)


def normalize_terrain_ids(
    terrain_ids: tuple[int, int, int] | None,
) -> tuple[TerrainTextureId, TerrainTextureId, TerrainTextureId]:
    if terrain_ids is None:
        return DEFAULT_TERRAIN_IDS
    try:
        return (
            TerrainTextureId(int(terrain_ids[0])),
            TerrainTextureId(int(terrain_ids[1])),
            TerrainTextureId(int(terrain_ids[2])),
        )
    except (TypeError, ValueError, IndexError):
        return DEFAULT_TERRAIN_IDS


class TerrainRuntime(msgspec.Struct):
    world_size: float = 1024.0
    render_resources: RenderResources = cast(RenderResources, None)

    def apply_bootstrap_terrain(
        self,
        *,
        terrain_ids: tuple[int, int, int],
        seed: int,
        layers: int = 3,
    ) -> None:
        base_id, overlay_id, detail_id = normalize_terrain_ids(terrain_ids)

        def _load(spec: tuple[str, str] | None) -> rl.Texture | None:
            if spec is None:
                return None
            key, rel_path = spec
            return self.render_resources.load_texture(
                str(key),
                cache_path=str(rel_path),
            )

        base = _load(terrain_texture_by_id(base_id))
        overlay = _load(terrain_texture_by_id(overlay_id))
        detail = _load(terrain_texture_by_id(detail_id)) or overlay or base

        if base is None and (base_id, overlay_id, detail_id) != DEFAULT_TERRAIN_IDS:
            # Fall back to default terrain if the chosen one is unavailable.
            base = _load(terrain_texture_by_id(TerrainTextureId.Q1_BASE))
            overlay = _load(terrain_texture_by_id(TerrainTextureId.Q1_OVERLAY))
            detail = _load(terrain_texture_by_id(TerrainTextureId.Q1_BASE)) or overlay or base

        if base is None:
            return

        self.render_resources.set_ground_textures(
            base=base,
            overlay=overlay,
            detail=detail,
        )
        self.render_resources.schedule_ground_generation(seed=int(seed), layers=int(layers))

    def set_terrain(
        self,
        *,
        base_key: str,
        overlay_key: str,
        base_path: str,
        overlay_path: str,
        detail_key: str | None = None,
        detail_path: str | None = None,
    ) -> None:
        base = self.render_resources.load_texture(base_key, cache_path=base_path)
        overlay = self.render_resources.load_texture(overlay_key, cache_path=overlay_path)
        detail = None
        if detail_key is not None and detail_path is not None:
            detail = self.render_resources.load_texture(detail_key, cache_path=detail_path)
        if detail is None:
            detail = overlay or base
        if base is None:
            return
        self.render_resources.set_ground_textures(
            base=base,
            overlay=overlay,
            detail=detail,
        )

    def schedule_from_rng_seed(self, *, seed: int, layers: int = 3) -> None:
        self.render_resources.schedule_ground_generation(seed=int(seed), layers=int(layers))

    def process_pending(self) -> None:
        self.render_resources.process_ground_pending()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import msgspec

from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rd, rl

from ...perks import PerkId
from ...perks.helpers import perk_active
from ...projectiles.types import ProjectileTemplateId
from ...sim.world_defs import KNOWN_PROJ_FRAMES
from ..projectile_draw import (
    ProjectileDrawCtx,
    SecondaryProjectileDrawCtx,
    draw_projectile_from_registry,
    draw_secondary_projectile_from_registry,
)
from ..projectile_render_registry import known_proj_rgb
from .context import WorldRenderCtx

if TYPE_CHECKING:
    from ...projectiles.types import Projectile, SecondaryProjectile
    from ..projectile_draw import ProjectileRendererLike
    from ..rtx.mode import RtxRenderMode


class _ProjectileRendererAdapter(msgspec.Struct):
    render_ctx: WorldRenderCtx
    camera: Vec2
    view_scale: Vec2

    @property
    def bullet_trail_texture(self) -> rl.Texture | None:
        return self.render_ctx.bullet_trail_texture

    @property
    def bullet_texture(self) -> rl.Texture | None:
        return self.render_ctx.bullet_texture

    @property
    def particles_texture(self) -> rl.Texture | None:
        return self.render_ctx.particles_texture

    @property
    def projs_texture(self) -> rl.Texture | None:
        return self.render_ctx.projs_texture

    @property
    def config(self):
        return self.render_ctx.config

    @property
    def players(self):
        return self.render_ctx.players

    @property
    def creatures(self):
        return self.render_ctx.creatures

    @property
    def elapsed_ms(self) -> float:
        return self.render_ctx.elapsed_ms

    @property
    def rtx_mode(self) -> "RtxRenderMode":
        return self.render_ctx.rtx_mode

    @staticmethod
    def _is_bullet_trail_type(type_id: int) -> bool:
        return is_bullet_trail_type(type_id)

    def world_to_screen(self, pos: Vec2) -> Vec2:
        return self.render_ctx._world_to_screen_with(pos, camera=self.camera, view_scale=self.view_scale)

    def _draw_bullet_trail(
        self,
        start: Vec2,
        end: Vec2,
        *,
        type_id: int,
        alpha: int,
        scale: float,
        angle: float,
    ) -> bool:
        return draw_bullet_trail(
            self.render_ctx,
            start,
            end,
            type_id=type_id,
            alpha=alpha,
            scale=scale,
            angle=angle,
        )

    @staticmethod
    def _bullet_sprite_size(type_id: int, *, scale: float) -> float:
        return bullet_sprite_size(type_id, scale=scale)

    def _draw_atlas_sprite(
        self,
        texture: rl.Texture,
        *,
        grid: int,
        frame: int,
        pos: Vec2,
        scale: float,
        rotation_rad: float = 0.0,
        tint: rl.Color = rl.WHITE,
    ) -> None:
        self.render_ctx._draw_atlas_sprite(
            texture,
            grid=grid,
            frame=frame,
            pos=pos,
            scale=scale,
            rotation_rad=rotation_rad,
            tint=tint,
        )


def draw_projectile(
    render_ctx: WorldRenderCtx,
    proj: Projectile,
    *,
    proj_index: int = 0,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    alpha: float = 1.0,
) -> None:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return

    texture = render_ctx.projs_texture
    type_id = proj.type_id
    proj_pos = proj.pos
    screen = render_ctx._world_to_screen_with(proj_pos, camera=camera, view_scale=view_scale)
    life = float(proj.life_timer)
    angle = float(proj.angle)

    adapter = _ProjectileRendererAdapter(render_ctx, camera=camera, view_scale=view_scale)
    registry_ctx = ProjectileDrawCtx(
        renderer=cast("ProjectileRendererLike", adapter),
        proj=proj,
        proj_index=int(proj_index),
        texture=texture,
        type_id=int(type_id),
        pos=proj_pos,
        screen_pos=screen,
        life=float(life),
        angle=float(angle),
        scale=float(scale),
        alpha=float(alpha),
    )
    if draw_projectile_from_registry(registry_ctx):
        return

    mapping = KNOWN_PROJ_FRAMES.get(type_id)
    if mapping is None:
        return
    if texture is None:
        if life < 0.39:
            return
        rl.draw_circle(
            int(screen.x),
            int(screen.y),
            max(1.0, 2.0 * scale),
            rl.Color(180, 180, 180, int(180 * alpha + 0.5)),
        )
        return

    grid, frame = mapping
    alpha_byte = int(clamp(clamp(life / 0.4, 0.0, 1.0) * 255.0 * alpha, 0.0, 255.0) + 0.5)
    red, green, blue = known_proj_rgb(type_id)
    tint = rl.Color(int(red), int(green), int(blue), alpha_byte)
    render_ctx._draw_atlas_sprite(
        texture,
        grid=grid,
        frame=frame,
        pos=screen,
        scale=0.6 * scale,
        rotation_rad=angle,
        tint=tint,
    )


def is_bullet_trail_type(type_id: int) -> bool:
    return 0 <= type_id < 8 or type_id == ProjectileTemplateId.SPLITTER_GUN


def bullet_sprite_size(type_id: int, *, scale: float) -> float:
    base = 4.0
    if type_id == ProjectileTemplateId.ASSAULT_RIFLE:
        base = 6.0
    elif type_id == ProjectileTemplateId.SUBMACHINE_GUN:
        base = 8.0
    return max(2.0, base * scale)


def draw_bullet_trail(
    render_ctx: WorldRenderCtx,
    start: Vec2,
    end: Vec2,
    *,
    type_id: int,
    alpha: int,
    scale: float,
    angle: float,
) -> bool:
    if render_ctx.bullet_trail_texture is None:
        return False
    if alpha <= 0:
        return False

    segment = end - start
    direction, dist = segment.normalized_with_length()

    # Native uses projectile travel direction as the side-offset basis and still emits the
    # trail quad even when origin=head (degenerate impact frames).
    if type_id in (ProjectileTemplateId.PISTOL, ProjectileTemplateId.ASSAULT_RIFLE):
        side_mul = 1.2
    elif type_id == ProjectileTemplateId.GAUSS_GUN:
        side_mul = 1.1
    else:
        side_mul = 0.7
    half = 1.5 * side_mul * scale

    if dist > 1e-6:
        side = direction.perp_left()
    else:
        side = Vec2.from_angle(angle)

    side_offset = side * half
    p0 = start - side_offset
    p1 = start + side_offset
    p2 = end + side_offset
    p3 = end - side_offset

    # Native uses additive blending for bullet trails and sets color slots per projectile type.
    # Gauss has a distinct blue tint; most other bullet trails are neutral gray.
    if type_id == ProjectileTemplateId.GAUSS_GUN:
        head_rgb = (51, 128, 255)  # (0.2, 0.5, 1.0)
    else:
        head_rgb = (128, 128, 128)  # (0.5, 0.5, 0.5)

    tail_rgb = (128, 128, 128)
    head = rl.Color(head_rgb[0], head_rgb[1], head_rgb[2], alpha)
    tail = rl.Color(tail_rgb[0], tail_rgb[1], tail_rgb[2], 0)

    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
    rl.rl_set_texture(render_ctx.bullet_trail_texture.id)
    rl.rl_begin(rd.RL_QUADS)
    rl.rl_color4ub(tail.r, tail.g, tail.b, tail.a)
    rl.rl_tex_coord2f(0.0, 0.0)
    rl.rl_vertex2f(p0.x, p0.y)
    rl.rl_color4ub(tail.r, tail.g, tail.b, tail.a)
    rl.rl_tex_coord2f(1.0, 0.0)
    rl.rl_vertex2f(p1.x, p1.y)
    rl.rl_color4ub(head.r, head.g, head.b, head.a)
    rl.rl_tex_coord2f(1.0, 0.5)
    rl.rl_vertex2f(p2.x, p2.y)
    rl.rl_color4ub(head.r, head.g, head.b, head.a)
    rl.rl_tex_coord2f(0.0, 0.5)
    rl.rl_vertex2f(p3.x, p3.y)
    rl.rl_end()
    rl.rl_set_texture(0)
    rl.end_blend_mode()
    return True


def draw_sharpshooter_laser_sight(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    alpha: float,
) -> None:
    """Laser sight overlay for the Sharpshooter perk (`projectile_render` @ 0x00422c70)."""

    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return
    if render_ctx.bullet_trail_texture is None:
        return

    players = render_ctx.players
    if not players:
        return

    tail_alpha = int(clamp(alpha * 0.5, 0.0, 1.0) * 255.0 + 0.5)
    head_alpha = int(clamp(alpha * 0.2, 0.0, 1.0) * 255.0 + 0.5)
    tail = rl.Color(255, 0, 0, tail_alpha)
    head = rl.Color(255, 0, 0, head_alpha)

    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
    rl.rl_set_texture(render_ctx.bullet_trail_texture.id)
    rl.rl_begin(rd.RL_QUADS)

    for player in players:
        if float(player.health) <= 0.0:
            continue
        if not perk_active(player, PerkId.SHARPSHOOTER):
            continue
        player_pos = player.pos

        aim_heading = float(player.aim_heading)
        aim_dir = Vec2.from_heading(aim_heading)
        start = player_pos + aim_dir * 15.0
        end = player_pos + aim_dir * 512.0

        start_screen = render_ctx._world_to_screen_with(start, camera=camera, view_scale=view_scale)
        end_screen = render_ctx._world_to_screen_with(end, camera=camera, view_scale=view_scale)
        segment = end_screen - start_screen
        direction, dist = segment.normalized_with_length()
        if dist <= 1e-3:
            continue

        thickness = max(1.0, 2.0 * scale)
        half = thickness * 0.5
        side_offset = direction.perp_left() * half
        p0 = start_screen - side_offset
        p1 = start_screen + side_offset
        p2 = end_screen + side_offset
        p3 = end_screen - side_offset

        rl.rl_color4ub(tail.r, tail.g, tail.b, tail.a)
        rl.rl_tex_coord2f(0.0, 0.0)
        rl.rl_vertex2f(p0.x, p0.y)
        rl.rl_color4ub(tail.r, tail.g, tail.b, tail.a)
        rl.rl_tex_coord2f(1.0, 0.0)
        rl.rl_vertex2f(p1.x, p1.y)
        rl.rl_color4ub(head.r, head.g, head.b, head.a)
        rl.rl_tex_coord2f(1.0, 0.5)
        rl.rl_vertex2f(p2.x, p2.y)
        rl.rl_color4ub(head.r, head.g, head.b, head.a)
        rl.rl_tex_coord2f(0.0, 0.5)
        rl.rl_vertex2f(p3.x, p3.y)

    rl.rl_end()
    rl.rl_set_texture(0)
    rl.end_blend_mode()


def draw_secondary_projectile(
    render_ctx: WorldRenderCtx,
    proj: SecondaryProjectile,
    *,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    alpha: float = 1.0,
) -> None:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return

    proj_pos = proj.pos
    screen = render_ctx._world_to_screen_with(proj_pos, camera=camera, view_scale=view_scale)
    proj_type = proj.type_id
    angle = float(proj.angle)

    adapter = _ProjectileRendererAdapter(render_ctx, camera=camera, view_scale=view_scale)
    registry_ctx = SecondaryProjectileDrawCtx(
        renderer=cast("ProjectileRendererLike", adapter),
        proj=proj,
        proj_type=proj_type,
        screen_pos=screen,
        angle=float(angle),
        scale=float(scale),
        alpha=float(alpha),
    )
    if draw_secondary_projectile_from_registry(registry_ctx):
        return

    rl.draw_circle(
        int(screen.x),
        int(screen.y),
        max(1.0, 4.0 * scale),
        rl.Color(200, 200, 220, int(200 * alpha + 0.5)),
    )



from __future__ import annotations

_NATIVE_FIND_SIZE_MARGIN_SCALE = 0.14285715
_NATIVE_FIND_SIZE_MARGIN_BIAS = 3.0


def native_find_size_margin(target_size: float) -> float:
    """Native collision threshold term used by `*_find_in_radius` routines."""

    return float(target_size) * _NATIVE_FIND_SIZE_MARGIN_SCALE + _NATIVE_FIND_SIZE_MARGIN_BIAS


__all__ = ["native_find_size_margin"]



from __future__ import annotations

from pathlib import Path
from typing import Literal

import typer

_SessionMode = Literal["survival", "rush", "quests"]
_ParsedNetcodeMode = Literal["rollback", "lockstep"]


def _parse_session_mode(mode: str) -> _SessionMode:
    mode_name = str(mode).strip().lower()
    if mode_name == "survival":
        return "survival"
    if mode_name == "rush":
        return "rush"
    if mode_name == "quests":
        return "quests"
    raise typer.BadParameter(
        f"unsupported mode {mode!r}; expected one of: survival, rush, quests",
        param_hint="--mode",
    )


def _parse_netcode_mode(raw: str) -> _ParsedNetcodeMode:
    value = str(raw).strip().lower()
    if value in {"rollback", "rb"}:
        return "rollback"
    if value == "lockstep":
        return "lockstep"
    raise typer.BadParameter(
        f"unsupported netcode {raw!r}; expected rollback|lockstep",
        param_hint="--netcode",
    )


def _run_game_with_pending_session(
    *,
    pending,
    base_dir: Path,
    assets_dir: Path | None,
    width: int | None,
    height: int | None,
    fps: int,
    debug: bool,
    rtx: bool,
) -> None:
    from ..game import GameConfig, run_game

    run_game(
        GameConfig(
            base_dir=base_dir,
            assets_dir=assets_dir,
            width=width,
            height=height,
            fps=fps,
            debug=bool(debug),
            rtx=bool(rtx),
            preserve_bugs=False,
            pending_network_session=pending,
        ),
    )



from __future__ import annotations

import warnings

from .types import Replay, current_replay_game_version


class ReplayGameVersionError(ValueError):
    """Raised when replay `game_version` is incompatible with runtime version."""


class ReplayGameVersionWarning(UserWarning):
    """Warned when replay and runtime build metadata differ while game version matches."""


def _split_game_version(value: str) -> tuple[str, str]:
    game_version, sep, build_metadata = str(value).partition("+")
    if not sep:
        return game_version.strip(), ""
    return game_version.strip(), build_metadata.strip()


def warn_on_game_version_mismatch(
    replay: Replay,
    *,
    action: str = "playback",
    current_version: str | None = None,
) -> None:
    """Require matching game version; warn when only build metadata differs."""

    expected = str(current_version).strip() if current_version is not None else str(current_replay_game_version()).strip()
    got = str(replay.header.game_version).strip()

    if not got:
        raise ReplayGameVersionError(
            f"Replay is missing game_version; {action} requires a matching game version (current={expected!r}).",
        )

    expected_game_version, _ = _split_game_version(expected)
    replay_game_version, _ = _split_game_version(got)
    if replay_game_version != expected_game_version:
        raise ReplayGameVersionError(
            f"Replay game_version mismatch; {action} requires a matching game version "
            f"(replay={got!r}, current={expected!r}).",
        )

    if got != expected:
        warnings.warn(
            "Replay game_version build metadata differs (git tag/hash mismatch); "
            "continuing because base game version matches "
            f"(replay={got!r}, current={expected!r}).",
            ReplayGameVersionWarning,
            stacklevel=2,
        )



from __future__ import annotations

from .registry import all_views, view_by_name


def _register_builtin_views() -> None:
    from . import arsenal_debug as _arsenal_debug  # noqa: F401
    from . import fonts as _fonts  # noqa: F401
    from . import game_over as _game_over  # noqa: F401
    from . import lighting_debug as _lighting_debug  # noqa: F401
    from . import perk_menu_debug as _perk_menu_debug  # noqa: F401
    from . import small_font_debug as _small_font_debug  # noqa: F401
    from . import spawn_plan as _spawn_plan  # noqa: F401


_register_builtin_views()

__all__ = ["all_views", "view_by_name"]



from __future__ import annotations

from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TypeAlias

from crimson.quest_level import QuestLevel
from grim.rand import CrtRand, RngTraceSink

from ...effects import FxQueue, FxQueueRotated
from ...game_modes import GameMode
from ...quests import quest_by_level
from ...quests.runtime import build_quest_spawn_table
from ...quests.types import QuestContext, QuestDefinition, SpawnEntry
from ...replay import Replay, apply_replay_bootstrap, warn_on_game_version_mismatch
from ...replay.checkpoints import ReplayCheckpoint
from ...replay.checkpoints import build_checkpoint as build_replay_checkpoint
from ...replay.header_settings import session_settings_from_replay_header
from ...replay.input_codec import unpack_tick_inputs
from ...weapon_runtime import weapon_assign_player
from ...weapons import WeaponId
from ..hooks import TickResult
from ..input_providers import ResolvedTick
from ..sessions import (
    DeterministicSession,
    QuestSpawnState,
    RushSpawnState,
    SurvivalSpawnState,
    quest_post_step,
    rush_input_transform,
    rush_mid_step,
    survival_mid_step,
)
from ..world_state import WorldState
from .replay_timing import should_apply_world_dt_steps_for_replay
from .setup import (
    ReplayRunnerError,
    RunResult,
    build_damage_scale_by_type,
    build_empty_fx_queues,
    player0_most_used_weapon_id,
    player0_shots,
    reset_players,
    status_from_snapshot,
)

RUSH_WEAPON_ID = WeaponId.ASSAULT_RIFLE
RUSH_FORCED_AMMO = 30.0

RngTraceDraw: TypeAlias = tuple[int, int, int]
TickRngTraceObserver: TypeAlias = Callable[[TickResult, tuple[RngTraceDraw, ...]], None]
TickProgressCallback: TypeAlias = Callable[[int], None]
TickBeginObserver: TypeAlias = Callable[
    [int, WorldState, float],
    None,
]
TickEndObserver: TypeAlias = Callable[[TickResult, WorldState], None]


def resolve_quest_level_from_replay(replay: Replay) -> str:
    quest_level = QuestLevel.try_parse(str(replay.header.quest_level))
    if quest_level is not None:
        return quest_level.to_string()

    # Legacy replays (e.g. capture-derived) may not encode the quest id. Classic quest RNG
    # seeding uses `major*100 + minor`, so we can often recover the level from `header.seed`.
    seed = int(replay.header.seed)
    major = seed // 100
    minor = seed % 100
    if 1 <= int(major) <= 5 and 1 <= int(minor) <= 10:
        return QuestLevel.from_parts(major, minor).to_string()
    return ""


def resolve_replay_quest_setup(
    replay: Replay,
    *,
    world_size: float,
    player_count: int,
) -> tuple[QuestDefinition, tuple[SpawnEntry, ...]]:
    quest_level = resolve_quest_level_from_replay(replay)
    quest = quest_by_level(quest_level) if quest_level else None
    if quest is None:
        raise ReplayRunnerError(f"unsupported quest replay: unknown quest_level={quest_level!r}")

    ctx = QuestContext(
        width=int(world_size),
        height=int(world_size),
        player_count=int(player_count),
    )
    spawn_entries = tuple(
        build_quest_spawn_table(
            quest,
            ctx,
            seed=int(replay.header.seed),
            hardcore=bool(replay.header.hardcore),
            full_version=True,
        ),
    )
    return quest, spawn_entries


def enforce_rush_loadout(world: WorldState) -> None:
    for player in world.players:
        if player.weapon.weapon_id != RUSH_WEAPON_ID:
            weapon_assign_player(player, RUSH_WEAPON_ID, state=world.state)
        # Native `rush_mode_update` forces assault rifle + 30 ammo every frame.
        player.weapon.ammo = float(RUSH_FORCED_AMMO)


@contextmanager
def _tick_rng_trace(rng: object, *, enabled: bool) -> Iterator[list[RngTraceDraw]]:
    draws: list[RngTraceDraw] = []
    if not enabled or not isinstance(rng, CrtRand):
        yield draws
        return

    previous_sink = rng.trace_sink

    def _sink(state_before_u32: int, state_after_u32: int, value_15: int) -> None:
        draws.append((int(state_before_u32), int(value_15), int(state_after_u32)))

    trace_sink: RngTraceSink = _sink
    rng.set_trace_sink(trace_sink)
    try:
        yield draws
    finally:
        rng.set_trace_sink(previous_sink)


@dataclass(slots=True, frozen=True)
class PlaybackWalkHooks:
    before_tick: TickBeginObserver | None = None
    after_tick: TickEndObserver | None = None
    on_progress: TickProgressCallback | None = None
    on_rng_trace: TickRngTraceObserver | None = None


@dataclass(slots=True, frozen=True)
class PlaybackWalkResult:
    start_tick: int
    next_tick_index: int
    ticks_completed: int


class PlaybackDriver:
    """Canonical replay driver."""

    def __init__(
        self,
        replay: Replay,
        *,
        max_ticks: int | None = None,
        trace_rng: bool = False,
        version_mismatch_action: str | None = "verification",
        world_size: float | None = None,
        fx_queue: FxQueue | None = None,
        fx_queue_rotated: FxQueueRotated | None = None,
        inter_tick_rand_draws: int = 0,
        inter_tick_rand_draws_by_tick: dict[int, int] | None = None,
        spawn_entries: tuple[SpawnEntry, ...] | None = None,
        quest_stage_major: int | None = None,
        quest_stage_minor: int | None = None,
        start_weapon_id: WeaponId | None = None,
    ) -> None:
        self.replay = replay
        self.max_ticks = max_ticks
        self.trace_rng = bool(trace_rng)
        self.inter_tick_rand_draws = max(0, int(inter_tick_rand_draws))
        self.inter_tick_rand_draws_by_tick = inter_tick_rand_draws_by_tick
        self._provided_world_size = float(world_size) if world_size is not None else None
        self._provided_fx_queue = fx_queue
        self._provided_fx_queue_rotated = fx_queue_rotated
        self._quest_spawn_entries = tuple(spawn_entries) if spawn_entries is not None else None
        self._quest_stage_major = quest_stage_major
        self._quest_stage_minor = quest_stage_minor
        self._quest_start_weapon_id = start_weapon_id
        self._quest_spawn_state: QuestSpawnState | None = None

        if version_mismatch_action is not None:
            warn_on_game_version_mismatch(replay, action=str(version_mismatch_action))

        self.session_settings = session_settings_from_replay_header(replay.header)

        mode_raw = int(self.session_settings.mode_id)
        try:
            self.mode_id = GameMode(mode_raw)
        except ValueError as exc:
            raise ReplayRunnerError(f"unsupported replay game_mode_id={mode_raw}") from exc

        tick_rate = int(self.session_settings.tick_rate)
        if tick_rate <= 0:
            raise ReplayRunnerError(f"invalid tick_rate: {tick_rate}")
        self.tick_rate = int(tick_rate)
        self.dt = 1.0 / float(self.tick_rate)

        self.world_size = self._resolve_world_size()
        self.world = self._prepare_world()
        self.fx_queue, self.fx_queue_rotated = self._resolve_fx_queues()

        apply_world_dt_steps = should_apply_world_dt_steps_for_replay(
            original_capture_replay=False,
        )
        self.session = self._build_session(apply_world_dt_steps=bool(apply_world_dt_steps))
        self._last_tick_rng_rows: tuple[RngTraceDraw, ...] = ()

        self.tick_limit = (
            len(replay.ticks)
            if self.max_ticks is None
            else min(len(replay.ticks), max(0, int(self.max_ticks)))
        )

    def _resolve_world_size(self) -> float:
        if self._provided_world_size is not None:
            return float(self._provided_world_size)
        return float(self.replay.header.world_size)

    def _prepare_world(self) -> WorldState:
        world = WorldState.build(
            world_size=float(self.world_size),
            demo_mode_active=False,
            hardcore=bool(self.replay.header.hardcore),
            quest_fail_retry_count=int(self.replay.header.quest_fail_retry_count),
            preserve_bugs=bool(self.session_settings.preserve_bugs),
        )
        reset_players(
            world.players,
            state=world.state,
            world_size=float(self.world_size),
            player_count=int(self.session_settings.player_count),
        )
        world.state.status = status_from_snapshot(
            quest_unlock_index=int(self.replay.header.status.quest_unlock_index),
            quest_unlock_index_full=int(self.replay.header.status.quest_unlock_index_full),
            weapon_usage_counts=self.replay.header.status.weapon_usage_counts,
        )

        if self.mode_id in (GameMode.SURVIVAL, GameMode.RUSH):
            apply_replay_bootstrap(
                self.replay.header,
                rng=world.state.rng,
                world_size=float(self.world_size),
            )
        else:
            world.state.rng.srand(int(self.replay.header.seed))

        return world

    def _resolve_fx_queues(self) -> tuple[FxQueue, FxQueueRotated]:
        if self._provided_fx_queue is not None and self._provided_fx_queue_rotated is not None:
            return self._provided_fx_queue, self._provided_fx_queue_rotated
        return build_empty_fx_queues()

    def _resolve_quest_setup(self) -> tuple[tuple[SpawnEntry, ...], int | None, int | None, WeaponId | None]:
        spawn_entries = self._quest_spawn_entries
        quest_stage_major = self._quest_stage_major
        quest_stage_minor = self._quest_stage_minor
        start_weapon_id = self._quest_start_weapon_id

        if spawn_entries is None:
            quest, spawn_entries = resolve_replay_quest_setup(
                self.replay,
                world_size=float(self.world_size),
                player_count=int(self.session_settings.player_count),
            )

            if quest_stage_major is None or quest_stage_minor is None:
                quest_stage_major, quest_stage_minor = quest.level_key
            if start_weapon_id is None:
                start_weapon_id = quest.start_weapon_id
        else:
            spawn_entries = tuple(spawn_entries)

        return spawn_entries, quest_stage_major, quest_stage_minor, start_weapon_id

    def _build_session(self, *, apply_world_dt_steps: bool) -> DeterministicSession:
        damage_scale_by_type = build_damage_scale_by_type()
        self._quest_spawn_state = None
        match self.mode_id:
            case GameMode.SURVIVAL:
                survival_spawn = SurvivalSpawnState()
                return DeterministicSession(
                    world=self.world,
                    world_size=float(self.world_size),
                    damage_scale_by_type=damage_scale_by_type,
                    fx_queue=self.fx_queue,
                    fx_queue_rotated=self.fx_queue_rotated,
                    game_mode=GameMode.SURVIVAL,
                    perk_progression_enabled=True,
                    detail_preset=int(self.replay.header.detail_preset),
                    gore_disabled=int(self.replay.header.gore_disabled),
                    game_tune_started=False,
                    apply_world_dt_steps=bool(apply_world_dt_steps),
                    clear_fx_queues_each_tick=True,
                    finalize_post_render_lifecycle=True,
                    mid_step_hook=lambda ctx: survival_mid_step(ctx, survival_spawn),
                )
            case GameMode.RUSH:
                enforce_rush_loadout(self.world)
                rush_spawn = RushSpawnState()
                return DeterministicSession(
                    world=self.world,
                    world_size=float(self.world_size),
                    damage_scale_by_type=damage_scale_by_type,
                    fx_queue=self.fx_queue,
                    fx_queue_rotated=self.fx_queue_rotated,
                    game_mode=GameMode.RUSH,
                    perk_progression_enabled=False,
                    detail_preset=int(self.replay.header.detail_preset),
                    gore_disabled=int(self.replay.header.gore_disabled),
                    game_tune_started=False,
                    clear_fx_queues_each_tick=True,
                    finalize_post_render_lifecycle=True,
                    elapsed_uses_raw_dt=True,
                    mid_step_hook=lambda ctx: rush_mid_step(ctx, rush_spawn),
                    before_step_hook=lambda: enforce_rush_loadout(self.world),
                    input_transform=rush_input_transform,
                )
            case GameMode.QUESTS:
                spawn_entries, quest_stage_major, quest_stage_minor, start_weapon_id = self._resolve_quest_setup()

                if quest_stage_major is not None and quest_stage_minor is not None:
                    self.world.state.quest_level = QuestLevel.from_parts(quest_stage_major, quest_stage_minor)

                weapon_id = start_weapon_id or WeaponId.PISTOL
                if weapon_id <= WeaponId.NONE:
                    weapon_id = WeaponId.PISTOL
                for player in self.world.players:
                    weapon_assign_player(player, weapon_id, state=self.world.state)

                self.world.creatures.capture_spawn_events_authoritative = False

                quest_state = QuestSpawnState(spawn_entries=tuple(spawn_entries))
                self._quest_spawn_state = quest_state

                return DeterministicSession(
                    world=self.world,
                    world_size=float(self.world_size),
                    damage_scale_by_type=damage_scale_by_type,
                    fx_queue=self.fx_queue,
                    fx_queue_rotated=self.fx_queue_rotated,
                    game_mode=GameMode.QUESTS,
                    perk_progression_enabled=True,
                    detail_preset=int(self.replay.header.detail_preset),
                    gore_disabled=int(self.replay.header.gore_disabled),
                    game_tune_started=False,
                    demo_mode_active=bool(self.world.state.demo_mode_active),
                    apply_world_dt_steps=bool(apply_world_dt_steps),
                    clear_fx_queues_each_tick=True,
                    finalize_post_render_lifecycle=True,
                    post_step_hook=lambda ctx: quest_post_step(ctx, quest_state),
                )
            case _:
                raise ReplayRunnerError(f"unsupported replay game_mode_id={int(self.mode_id)}")

    def _prepare_tick(self, *, tick_index: int) -> None:
        if int(tick_index) < 0 or int(tick_index) >= int(self.tick_limit):
            raise ReplayRunnerError(f"tick_index out of range: {tick_index} (tick_limit={self.tick_limit})")

        state = self.world.state
        state.game_mode = self.mode_id
        state.demo_mode_active = False

        if self.inter_tick_rand_draws_by_tick is not None:
            draws = self.inter_tick_rand_draws_by_tick.get(int(tick_index))
            if draws is None:
                draws = int(self.inter_tick_rand_draws)
            for _ in range(max(0, int(draws))):
                state.rng.rand()

    def build_checkpoint(
        self,
        *,
        tick_result: TickResult,
        use_world_step_creature_count: bool = False,
    ) -> ReplayCheckpoint:
        return build_replay_checkpoint(
            tick_index=int(tick_result.source_tick.tick_index),
            world=self.world,
            elapsed_ms=float(self.elapsed_ms),
            creature_count_override=(
                int(tick_result.payload.creature_count_world_step)
                if bool(use_world_step_creature_count)
                else None
            ),
            rng_marks=dict(tick_result.payload.rng_marks),
            deaths=tick_result.payload.step.events.deaths,
            events=tick_result.payload.step.events,
        )

    def step_tick(self, tick_index: int) -> TickResult:
        tick_index = int(tick_index)
        self._prepare_tick(tick_index=tick_index)
        self._last_tick_rng_rows = ()
        replay_tick = self.replay.ticks[tick_index]
        dt_tick = float(replay_tick.dt)
        inputs = unpack_tick_inputs(replay_tick.inputs)
        commands = list(replay_tick.commands)
        with _tick_rng_trace(self.world.state.rng, enabled=bool(self.trace_rng)) as tick_rng_rows:
            source_tick = ResolvedTick(
                tick_index=int(tick_index),
                dt_seconds=float(dt_tick),
                inputs=list(inputs),
                commands=list(commands),
            )

            timing = self.session.timing_for_dt(dt_tick)
            session_tick = self.session.step_tick(
                timing=timing,
                inputs=inputs,
                trace_rng=self.trace_rng,
                commands=commands,
            )
            tick_result = TickResult(
                source_tick=source_tick,
                payload=session_tick,
                replay_tick_index=int(tick_index),
            )

        if self.inter_tick_rand_draws_by_tick is None:
            draws = max(0, int(self.inter_tick_rand_draws))
            for _ in range(draws):
                self.world.state.rng.rand()
        self._last_tick_rng_rows = tuple(tick_rng_rows)
        return tick_result

    def walk_ticks(
        self,
        *,
        start_tick: int = 0,
        stop_tick: int | None = None,
        hooks: PlaybackWalkHooks | None = None,
    ) -> PlaybackWalkResult:
        requested_start_tick = int(start_tick)
        if requested_start_tick < 0:
            raise ReplayRunnerError(f"invalid start_tick: {requested_start_tick}")
        requested_stop_tick = int(self.tick_limit) if stop_tick is None else int(stop_tick)
        if requested_stop_tick < requested_start_tick:
            raise ReplayRunnerError(
                f"invalid tick range: start_tick={requested_start_tick} stop_tick={requested_stop_tick}",
            )

        tick_limit = int(self.tick_limit)
        next_tick_index = min(requested_start_tick, tick_limit)
        stop_tick_index = min(requested_stop_tick, tick_limit)
        active_hooks = hooks if hooks is not None else PlaybackWalkHooks()

        while next_tick_index < stop_tick_index:
            if active_hooks.before_tick is not None:
                active_hooks.before_tick(
                    int(next_tick_index),
                    self.world,
                    float(self.replay.ticks[next_tick_index].dt),
                )
            tick_result = self.step_tick(next_tick_index)
            next_tick_index = int(tick_result.source_tick.tick_index) + 1

            if active_hooks.after_tick is not None:
                active_hooks.after_tick(tick_result, self.world)
            if active_hooks.on_rng_trace is not None:
                active_hooks.on_rng_trace(tick_result, self._last_tick_rng_rows)
            if active_hooks.on_progress is not None:
                active_hooks.on_progress(int(next_tick_index))

        return PlaybackWalkResult(
            start_tick=min(requested_start_tick, tick_limit),
            next_tick_index=int(next_tick_index),
            ticks_completed=int(next_tick_index - min(requested_start_tick, tick_limit)),
        )

    def run(
        self,
        *,
        hooks: PlaybackWalkHooks | None = None,
    ) -> RunResult:
        self.walk_ticks(
            start_tick=0,
            stop_tick=int(self.tick_limit),
            hooks=hooks,
        )
        return self.build_run_result(ticks=int(self.tick_limit))

    def build_run_result(self, *, ticks: int) -> RunResult:
        shots_fired, shots_hit = player0_shots(self.world.state)
        most_used_weapon_id = player0_most_used_weapon_id(self.world.state, self.world.players)
        score_xp = int(self.world.players[0].experience) if self.world.players else 0

        return RunResult(
            game_mode_id=self.mode_id,
            tick_rate=int(self.tick_rate),
            ticks=int(ticks),
            elapsed_ms=int(self.elapsed_ms),
            score_xp=int(score_xp),
            creature_kill_count=int(self.world.creatures.kill_count),
            most_used_weapon_id=most_used_weapon_id,
            shots_fired=int(shots_fired),
            shots_hit=int(shots_hit),
            rng_state=int(self.world.state.rng.state),
        )

    @property
    def elapsed_ms(self) -> float:
        quest_state = self._quest_spawn_state
        if quest_state is not None:
            return float(quest_state.spawn_timeline_ms)
        return float(self.session.elapsed_ms)

    @property
    def quest_spawn_state(self) -> QuestSpawnState | None:
        return self._quest_spawn_state


def build_verify_playback_driver(
    replay: Replay,
    *,
    max_ticks: int | None = None,
    warn_on_version_mismatch: bool = True,
    trace_rng: bool = False,
    inter_tick_rand_draws: int = 0,
    inter_tick_rand_draws_by_tick: dict[int, int] | None = None,
    spawn_entries: tuple[SpawnEntry, ...] | None = None,
    quest_stage_major: int | None = None,
    quest_stage_minor: int | None = None,
    start_weapon_id: WeaponId | None = None,
) -> PlaybackDriver:
    """Build the canonical headless/verification replay driver."""

    return PlaybackDriver(
        replay,
        max_ticks=max_ticks,
        trace_rng=bool(trace_rng),
        version_mismatch_action=("verification" if bool(warn_on_version_mismatch) else None),
        inter_tick_rand_draws=int(inter_tick_rand_draws),
        inter_tick_rand_draws_by_tick=inter_tick_rand_draws_by_tick,
        spawn_entries=spawn_entries,
        quest_stage_major=quest_stage_major,
        quest_stage_minor=quest_stage_minor,
        start_weapon_id=start_weapon_id,
    )


def build_runtime_playback_driver(
    replay: Replay,
    *,
    max_ticks: int | None,
    trace_rng: bool,
    world_size: float,
    fx_queue: FxQueue,
    fx_queue_rotated: FxQueueRotated,
    spawn_entries: tuple[SpawnEntry, ...] | None = None,
    quest_stage_major: int | None = None,
    quest_stage_minor: int | None = None,
    start_weapon_id: WeaponId | None = None,
) -> PlaybackDriver:
    """Build the canonical live replay-playback driver."""

    return PlaybackDriver(
        replay,
        max_ticks=max_ticks,
        trace_rng=bool(trace_rng),
        version_mismatch_action=None,
        world_size=float(world_size),
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        spawn_entries=spawn_entries,
        quest_stage_major=quest_stage_major,
        quest_stage_minor=quest_stage_minor,
        start_weapon_id=start_weapon_id,
    )



from __future__ import annotations

import msgspec

from .ids import BonusId


class BonusDurationPayload(msgspec.Struct, frozen=True):
    seconds: int


class BonusWeaponPayload(msgspec.Struct, frozen=True):
    weapon_id: int


class BonusPointsPayload(msgspec.Struct, frozen=True):
    points: int


BonusPayload = BonusDurationPayload | BonusWeaponPayload | BonusPointsPayload


def bonus_duration_payload(seconds: int) -> BonusDurationPayload:
    return BonusDurationPayload(seconds=int(seconds))


def bonus_weapon_payload(weapon_id: int) -> BonusWeaponPayload:
    return BonusWeaponPayload(weapon_id=int(weapon_id))


def bonus_points_payload(points: int) -> BonusPointsPayload:
    return BonusPointsPayload(points=int(points))


def bonus_payload_from_bonus(*, bonus_id: BonusId, amount: int) -> BonusPayload:
    amount = int(amount)
    if bonus_id == BonusId.WEAPON:
        return bonus_weapon_payload(amount)
    if bonus_id == BonusId.POINTS:
        return bonus_points_payload(amount)
    return bonus_duration_payload(amount)


def bonus_payload_value(payload: BonusPayload) -> int:
    if isinstance(payload, BonusDurationPayload):
        return int(payload.seconds)
    if isinstance(payload, BonusWeaponPayload):
        return int(payload.weapon_id)
    return int(payload.points)


__all__ = [
    "BonusDurationPayload",
    "BonusPayload",
    "BonusPointsPayload",
    "BonusWeaponPayload",
    "bonus_duration_payload",
    "bonus_payload_from_bonus",
    "bonus_payload_value",
    "bonus_points_payload",
    "bonus_weapon_payload",
]
